#ifndef ATIIXP_H
#define ATIIXP_H

#include <linux/config.h>
#include <linux/pci.h>
#include <linux/ide.h>

#define ATIIXP_DEBUG_DRIVE_INFO		1

#define DISPLAY_ATIIXP_TIMINGS

#if defined(DISPLAY_ATIIXP_TIMINGS) && defined(CONFIG_PROC_FS)
#include <linux/stat.h>
#include <linux/proc_fs.h>

static u8 atiixp_proc;

static int atiixp_get_info(char *, char **, off_t, int);

static ide_pci_host_proc_t atiixp_procs[] __devinitdata = {
	{
		.name		= "atiixp",
		.set		= 1,
		.get_info	= atiixp_get_info,
		.parent		= NULL,
	},
};
#endif  /* defined(DISPLAY_ATIIXP_TIMINGS) && defined(CONFIG_PROC_FS) */

static void init_setup_atiixp(struct pci_dev *, ide_pci_device_t *);
static unsigned int __devinit init_chipset_atiixp(struct pci_dev *, const char *);
static void init_hwif_atiixp(ide_hwif_t *);
static void init_dma_atiixp(ide_hwif_t *, unsigned long);


/*
 *	Table of the various ATIIXP capability blocks
 *
 */
 
static ide_pci_device_t atiixp_pci_info[] __devinitdata = {
	{	/* 0 */
		.vendor		= PCI_VENDOR_ID_ATI,
		.device		= PCI_DEVICE_ID_ATI_IXP_IDE,
		.name		= "ATIIXP",
		.init_setup	= init_setup_atiixp,
		.init_chipset	= init_chipset_atiixp,
		.init_iops	= NULL,
		.init_hwif	= init_hwif_atiixp,
		.init_dma	= init_dma_atiixp,
		.channels	= 2,
		.autodma	= AUTODMA,
		.enablebits	= {{0x48,0x01,0x00}, {0x48,0x08,0x00}},
		.bootable	= ON_BOARD,
		.extra		= 0,
	}
};

#define ATIIXP_IDE_PIO_TIMING		0x40
#define ATIIXP_IDE_MDMA_TIMING		0x44
#define ATIIXP_IDE_PIO_CONTROL		0x48
#define ATIIXP_IDE_PIO_MODE		0x4a
#define ATIIXP_IDE_UDMA_CONTROL		0x54
#define ATIIXP_IDE_UDMA_MODE		0x56

typedef struct {
  u8 command_width;
    u8 recover_width;
} atiixp_ide_timing;

static atiixp_ide_timing pio_timing[] = {
  {0x05, 0x0d},
  {0x04, 0x07},
  {0x03, 0x04},
  {0x02, 0x02},
  {0x02, 0x00},
};
static atiixp_ide_timing mdma_timing[] = {
  {0x07, 0x07},
  {0x02, 0x01},
  {0x02, 0x00},
};

#endif /* ATIIXP_H */
