// ****************************************************************************
//
//              CDaffyDuck.H
//
//              Include file for interfacing with the CDaffyDuck class.
//
//              A daffy duck maintains a scatter-gather list used by the DSP to 
//              transfer audio data via bus mastering.
//
//              The scatter-gather list takes the form of a circular buffer of
//              duck entries; each duck entry is an address/count pair that points
//              to an audio data buffer somewhere in memory.  
//
//              Although the scatter-gather list itself is a circular buffer, that
//              does not mean that the audio data pointed to by the scatter-gather
//              list is necessarily a circular buffer.  The audio buffers pointed to
//              by the SG list may be either a non-circular linked list of buffers
//              or a ring buffer.
//
//              If you want a ring DMA buffer for your audio data, refer to the 
//              Wrap() method, below.
//
//              The name "daffy duck" is an inside joke that dates back to the 
//              original VxD for Windows 95.
//
//              Set editor tabs to 3 for your viewing pleasure.
//
//---------------------------------------------------------------------------
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************


#ifndef _SGLIST_H_
#define _SGLIST_H_

#ifdef _DEBUG
#define INTEGRITY_CHECK
#endif

//
// sg_entry is a single entry for the scatter-gather list.  The array of struct
// sg_entry struct is read by the DSP, so all values must be little-endian.
//

typedef struct {
  u32 PhysAddr;
  u32 Size;
} sg_entry;

#define MAX_SGLIST_ENTRIES 500

#endif // _SGLIST_H_
