// ****************************************************************************
//
//		CEchoGals.H
//
//		Include file for the CEchoGals generic driver class.
//		Set editor tabs to 3 for your viewing pleasure.
//
// 	CEchoGals is the big daddy class of the generic code.  It is the upper
//		edge of the generic code - that is, it is the interface between the
// 	operating system-specific code and the generic driver.
//
//		There are a number of terms in this file that won't make any sense unless
// 	you go read EchoGalsXface.h first.
//
// ----------------------------------------------------------------------------
//
//		Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//		All rights reserved
//		www.echoaudio.com
//
//		Permission is hereby granted, free of charge, to any person obtaining a
//		copy of this software and associated documentation files (the
//		"Software"), to deal with the Software without restriction, including
//		without limitation the rights to use, copy, modify, merge, publish,
//		distribute, sublicense, and/or sell copies of the Software, and to
//		permit persons to whom the Software is furnished to do so, subject to
//		the following conditions:
//
//		- Redistributions of source code must retain the above copyright
//		notice, this list of conditions and the following disclaimers.
//
//		- Redistributions in binary form must reproduce the above copyright
//		notice, this list of conditions and the following disclaimers in the
//		documentation and/or other materials provided with the distribution.
//
//		- Neither the name of Echo Digital Audio, nor the names of its
//		contributors may be used to endorse or promote products derived from
//		this Software without specific prior written permission.
//
//		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//		EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//		MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//		IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//		ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//		TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//		SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#if !defined(ECHO24_FAMILY) && !defined(ECHOGALS_FAMILY)
#error No family defined !
#endif

//	Prevent problems with multiple includes
#ifndef _ECHOAUDIO_H_
#define _ECHOAUDIO_H_

#ifdef _DEBUG
#ifdef ECHO_WDM
#pragma optimize("",off)
#endif
#endif


//
// Each project creates this file to support the OS it is targeted for
//
#include "OsSupport_types.h"

// Include the definitions for the card family
#include "Echo_family.h"

#undef MIDI_SUPPORT

// Interface definitions

#include "EchoGalsXface.h"
#include "MixerXface.h"
#include "Echo_Dsp.h"


//
//	Pipe states
//
#define	PIPE_STATE_RESET		0	// Pipe has been reset
#define	PIPE_STATE_STOPPED		1	// Pipe has been stopped
#define	PIPE_STATE_STARTED		2	// Pipe has been started
#define	PIPE_STATE_PENDING		3	// Pipe has pending start


//
// Prototypes to make the Mac compiler happy
//
//ECHOSTATUS CheckSetting(INT32 iValue,INT32 iMin,INT32 iMax);
//INT32 PanToDb( INT32 iPan );


struct AudioPipe {
  u16 Channel;
  u16 Interleave;
  char IsInput;
  char IsCyclic;
  char State;
  u8 BytesPerSample;
  sg_entry *sglist;
  dma_addr_t sglist_phys;
  int sgHead;
};



struct CEchoGals {
  u16 Flags;                             // See ECHOGALS_FLAG_??? flags defined in EchoGalsXface.h
  u16 DeviceId, SubDeviceId, CardType;
  void *m_pvSharedMemory;                // Shared memory addr assigned by PNP
  u32 AudioOpenMask;                     // Audio channels open mask
  u32 AudioCyclicMask;                   // Audio use cyclic buffers mask

  u32 m_dwSampleRate;                    // Card sample rate in Hz
  struct AudioPipe Pipes[ECHO_MAXAUDIOPIPES];
  char MetersOn;

  u16 m_dwMidiInOpen;	                 // Midi in channels open mask
  u16 m_dwMidiOutOpen;	                 // Midi out channels open mask
  void *ossprivate_data;                 // User defined data. In the Linux driver it contains the *pci used to allocate dma-able data for the given card

  char m_fMixerDisabled;
  u8 DigitalMode;                        // Current Digital mode (see DIGITAL_MODE_*)
  u8 ProfessionalSpdif;                  // Gina24, Mona, Layla24 - only
  u8 SpdifStatus;                        // Gina20, Darla20, Darla24 - only
  u8 ClockState;                         // Gina20, Darla20, Darla24 - only
  u8 InputClock;
  u8 OutputClock;
  char NominalLevel[ECHO_MAXAUDIOPIPES]; // True == -10dBV  False == +4dBu
  int InLevel[ECHO_MAXAUDIOINPUTS];      // Input level -25..+25 dB  24.8 fixed point
  int OutLevel[ECHO_MAXAUDIOOUTPUTS];    // Output level -128..+6 dB  24.8 fixed point  (-128==muted)
  int MonitorLevel[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOINPUTS];  // -25..+25
  int VmixerLevel[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOOUTPUTS];  // -25..+25

  ////// Dsp stuff

  volatile PDspCommPage m_pDspCommPage;  // Physical memory seen by DSP

  // These members are not seen by the DSP; they are used internally by this class.
  u16 m_wNumPipesOut;
  u16 m_wNumPipesIn;
  u16 m_wNumBussesOut;
  u16 m_wNumBussesIn;
  u16 m_wFirstDigitalBusOut;
  u16 m_wFirstDigitalBusIn;
  u16 m_wNumMidiOut;                     // # MIDI out channels
  u16 m_wNumMidiIn;                      // # MIDI in  channels
  u32 DigitalModes;                      // Bitmask of supported modes (see ECHOCAPS_HAS_DIGITAL_MODE_*)

  char m_fHasVmixer;
  char m_bHasASIC;                       // Set TRUE if card has an ASIC
  char m_bASICLoaded;                    // Set TRUE when ASIC loaded
  char m_bBadBoard;                      // Set TRUE if DSP won't load

  u16 *m_pwDspCode;                      // Current DSP code loaded, NULL if nothing loaded
  u16 *m_pwDspCodeToLoad;                // DSP code to load
  u32 m_dwCommPagePhys;                  // Physical addr of this object
  volatile u32 *m_pdwDspRegBase;         // DSP's register base
  u32 ActiveMask;                        // Chs. active mask or punks out

  u8 *m_pbyAsic;                         // Current ASIC code

  // attempted to load the DSP & ASIC
#ifdef DIGITAL_INPUT_AUTO_MUTE_SUPPORT
  char m_fDigitalInAutoMute;
#endif

#ifdef MIDI_SUPPORT
  u16 m_wMidiOnCount;                    // Count MIDI enabled cmds
  u64 m_ullMidiInTime;                   // Last time MIDI in occured
  u64 m_ullMidiOutTime;                  // Last time MIDI out occured
  u64 m_ullNextMidiWriteTime;            // Next time to try MIDI output
  u16 m_wMtcState;                       // State for MIDI input parsing state machine
#endif
};






// Constructor and destructor
// (Note that you must call InitHw before you can do anything useful)
struct CEchoGals *NewEchoGals(u16, u16);
void DelEchoGals(struct CEchoGals *);

// Init the line level classes.
// This MUST be called from within any derived classes as part of
// InitHw after the DSP is up and running!
int InitLineLevels(struct CEchoGals *eg);

// Initialize the hardware, create the CDspCommObject class, return status.
int InitHw(struct CEchoGals *eg);

// This is called from within an interrupt handler.  It starts interrupt
// handling.  Returns error status if this interrupt is not ours.
// Returns ECHOSTATUS_OK if it is.
//
// OS dependent code handles routing to this point.
int ServiceIrq(struct CEchoGals *eg, int *fMidiReceived);

// Call this to find out if this card can handle a given sample rate
int QueryAudioSampleRate(struct CEchoGals *eg, u32 dwSampleRate);

// Return the capabilities of this card; card type, card name,
// # analog inputs, # analog outputs, # digital channels,
// # MIDI ports and supported clocks.
// See ECHOGALS_CAPS definition above.
int GetCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities);

// GetBaseCapabilities is called by GetCapabilities to initialize Capabilities structure with defaults.
int GetBaseCapabilities(struct CEchoGals *eg, PECHOGALS_CAPS pCapabilities);

int SetOutputGain(struct CEchoGals *eg, u16 Channel, int Gain);
int SetInputGain(struct CEchoGals *eg, u16 Channel, int Gain);
int SetMonitorGain(struct CEchoGals *eg, u16 OutChannel, u16 InChannel, int Gain);
int SetVmixerGain(struct CEchoGals *eg, u16 Pipe, u16 Output, int Gain);

int QueryAudioSampleRate(struct CEchoGals *eg, u32 SampleRate);
int SetAudioSampleRate(struct CEchoGals *eg, u32 Rate);
int QueryAudioFormat(struct CEchoGals *eg, u16 PipeIndex, ECHOGALS_AUDIOFORMAT *AudioFormat);
int SetAudioFormat(struct CEchoGals *eg, u16 PipeIndex, ECHOGALS_AUDIOFORMAT *AudioFormat);
int SetProfessionalSpdif(struct CEchoGals *eg, char Prof);
int SetDigitalMode(struct CEchoGals *eg, u8 Mode);
int SetNominalLevel(struct CEchoGals *eg, u16 Index, char Consumer);
int SetInputClock(struct CEchoGals *eg, u16 Clock);
int SetOutputClock(struct CEchoGals *eg, u16 Clock);
void SetFlags(struct CEchoGals *eg, u16 Flags);
void ClearFlags(struct CEchoGals *eg, u16 Flags);
void SetMetersOn(struct CEchoGals *eg, char On);
//void GetAudioMeters(struct CEchoGals *eg, ECHOGALS_METERS *Meters);
void GetAudioMeters(struct CEchoGals *eg, long *Meters);

int OpenAudio(struct CEchoGals *eg, u16 Channel, u16 Interleave, char IsInput, u16 *pipeindex);
int VerifyAudioOpen(struct CEchoGals *eg, u16 PipeIndex);
int CloseAudio(struct CEchoGals *eg, u16 PipeIndex);
int StartAudio(struct CEchoGals *eg, u16 PipeIndex);
int StartAudioMask(struct CEchoGals *eg,  u32 Mask);
int StopAudio(struct CEchoGals *eg, u16 PipeIndex);
int StopAudioMask(struct CEchoGals *eg, u32 Mask);
int ResetAudio(struct CEchoGals *eg, u16 PipeIndex);
int ResetAudioMask(struct CEchoGals *eg, u32 Mask);

int AllocateSGList(struct CEchoGals *eg, u16 PipeIndex);
void FreeSGList(struct CEchoGals *eg, u16 PipeIndex);
int SGListAddMapping(struct CEchoGals *eg, u16 PipeIndex, dma_addr_t Address, size_t Length, char Irq);
int SGListWrap(struct CEchoGals *eg, u16 PipeIndex);

int GetAudioPositionPtr(struct CEchoGals *eg, u16 PipeIndex, u32 **Position);






////// Dsp funcs


int NewDspComm(struct CEchoGals *eg);
void CreateCommPage(struct CEchoGals *eg, u32 *DspRegBase);
void DeleteCDspComm(struct CEchoGals *eg);
int GoComatose(struct CEchoGals *eg);
int LoadFirmware(struct CEchoGals *eg);
int LoadDSP(struct CEchoGals *eg, u16 * pCode);
int LoadASIC(struct CEchoGals *eg);
int LoadASICgeneric(struct CEchoGals *eg, u32 dwCmd, u8 * pCode, u32 dwSize);
int CheckAsicStatus(struct CEchoGals *eg);
void RestoreDspSettings(struct CEchoGals *eg);

int WaitForHandshake(struct CEchoGals *eg);
void ClearHandshake(struct CEchoGals *eg);
int SendVector(struct CEchoGals *eg, u32 Command);
void SetControlRegister(struct CEchoGals *eg, u32 ontrolRegister);
int Read_DSP(struct CEchoGals *eg, u32 * pdwData);
int Write_DSP(struct CEchoGals *eg, u32 dwData);
u32 GetStatusReg(struct CEchoGals *eg);
void AckInt(struct CEchoGals *eg);

#ifdef DSP_56361
int InstallResidentLoader(struct CEchoGals *eg);
#endif

// Gina24, Mona, Layla24
int WriteControlReg(struct CEchoGals *eg, u32 dwControlReg);
u32 GetControlRegister(struct CEchoGals *eg);


u32 GetSampleRate(const struct CEchoGals *eg);
u32 SetSampleRate(struct CEchoGals *eg, u32 NewSampleRate);
u32 GetInputClockDetect(const struct CEchoGals *eg);
int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode);
void DspSetFlags(struct CEchoGals *eg, u32 Flags);
void DspClearFlags(struct CEchoGals *eg, u32 Flags);

int UpdateOutputLineLevel(struct CEchoGals *eg);
int UpdateInputLineLevel(struct CEchoGals *eg);
int UpdateVmixerLevel(struct CEchoGals *eg);
int UpdateFlags(struct CEchoGals *eg);

int DspSetFormat(struct CEchoGals *eg, u16 Index, ECHOGALS_AUDIOFORMAT *Format, char DitherDigitalInputs);

int StartTransport(struct CEchoGals *eg, u32 ChannelMask, u32 CyclicMask);
int StopTransport(struct CEchoGals *eg, u32 ChannelMask);
int ResetTransport(struct CEchoGals *eg, u32 ChannelMask);


#endif // _ECHOAUDIO_H_

