/*
 *  ALSA mixer console for Echoaudio soundcards.
 *  Copyright (C) 2003 Giuliano Pochini <pochini@shiny.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define VERSION "Echoaudio console v0.8.3"


/*
 This option changes the mixer so that you can select the output channel on
 the mixer in place of the input channel.
*/
//#define REVERSE



#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <alsa/asoundlib.h>

#define DSP_TO_GENERIC(Value) (Value << 8)

#define LEVELSTEP DSP_TO_GENERIC(1)                     // 1dB
#define MAXMODES 16 //@fare dinamico
#define INPUT 0
#define OUTPUT 1
#define ECHO_MAXAUDIO_IOS 32                            // It's the maximum between inputs an outputs
#define ECHO_MAXAUDIOINPUTS     32                      // Max audio input channels
#define ECHO_MAXAUDIOOUTPUTS    32                      // Max audio output channels
#define ECHOGAIN_MUTED          DSP_TO_GENERIC(-128)    // Minimum possible gain
#define ECHOGAIN_MINOUT         DSP_TO_GENERIC(-128)    // Min output gain in dB
#define ECHOGAIN_MAXOUT         DSP_TO_GENERIC(6)       // Max output gain in dB
#define ECHOGAIN_MININP         DSP_TO_GENERIC(-25)     // Min input gain in dB
#define ECHOGAIN_MAXINP         DSP_TO_GENERIC(25)      // Max input gain in dB

#define BARWIDTH 5   // Width of VU-meters bars
#define BARSEP 2     // Spacing between VU-meters bars
#define BORDER 6     // Inner window border
#define SPACING 8    // Spacing of control sections

char card[64];
char dmodeName[MAXMODES][64], clocksrcName[MAXMODES][64], spdifmodeName[MAXMODES][64];
int nLOut, nIn, fdIn, fdOut, nPOut;
int ndmodes, nclocksrc, nspdifmodes;
int GMixerRow, InputChannel, OutputChannel, Gang;
int involId, outvolId, masterId, vmixerId, p4InId, p4OutId, dmodeId, clocksrcId, spdifmodeId, vuswitchId, vumetersId, channelsId, ditherId;
int mixerInputs, mixerOutputs, vmixerPipes, vmixerOutputs;      // Mixer dimensions. They should match the number of channels.
int metersStream, metersNumber, metersTypes;
int outvolCount;
int mouseY, mouseButton;
int dmodeVal, clocksrcVal, spdifmodeVal;
int VUwidth, VUheight, Mixwidth, Mixheight;

#define NOPOS 999999
struct geometry {
  int st;       // window status: 0 = hidden ; 1 = visible ; NOPOS = no setting
  int x,y;
  int w,h;
} Mainw_geom, PVw_geom, LVw_geom, Mixerw_geom, VUw_geom, GMw_geom;

// This structure contains the last row of each section of the graphic mixer window
struct {
  int Monitor;
  int Vmixer;
  int Line;
} GMixerSection;

struct mixel {
  int id;
  int val;
} Mixer[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOINPUTS], Vmixer[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOOUTPUTS];
snd_ctl_t *ctlhandle;
int LineOutVol[ECHO_MAXAUDIOOUTPUTS];
int LineInVol[ECHO_MAXAUDIOINPUTS];
int PcmOutVol[ECHO_MAXAUDIOOUTPUTS];
int PcmOutVol[ECHO_MAXAUDIOOUTPUTS];

GtkWidget *invol[ECHO_MAXAUDIOOUTPUTS], *invol_lbl[ECHO_MAXAUDIOOUTPUTS];       // PCM in
GtkObject *involadj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *outvol[ECHO_MAXAUDIOOUTPUTS], *outvol_lbl[ECHO_MAXAUDIOOUTPUTS];     // PCM out
GtkObject *outvoladj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *mstvol[ECHO_MAXAUDIOOUTPUTS], *mstvol_lbl[ECHO_MAXAUDIOOUTPUTS];     // Master out
GtkObject *mstadj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *mixvol[ECHO_MAXAUDIOOUTPUTS], *mixvol_lbl[ECHO_MAXAUDIOOUTPUTS];     // Mixer
GtkObject *mixeradj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *vmixervol[ECHO_MAXAUDIOOUTPUTS], *vmixer_lbl[ECHO_MAXAUDIOOUTPUTS];  // Vmixer
GtkObject *vmixeradj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *p4dbuOut[ECHO_MAXAUDIOOUTPUTS], *p4dbuIn[ECHO_MAXAUDIOINPUTS];       // +4dBu/-10dBV toggles
GtkWidget *mixin[ECHO_MAXAUDIOINPUTS], *mixout[ECHO_MAXAUDIOOUTPUTS];           // In and out selectors
GtkWidget *dmodeOpt, *clocksrcOpt, *spdifmodeOpt, *ditherToggle;
GtkWidget *window, *Mainwindow, *PVwindow, *LVwindow, *Mixerwindow, *VUwindow, *GMwindow;
GtkWidget *VUs[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *VUdarea, *Mixdarea;
gint VUtimer, Mixtimer;

GdkGC *gc=0;
static GdkPixmap *VUpixmap = NULL;
static GdkPixmap *Mixpixmap = NULL;
GdkFont *fnt;


int CountBits(int n) {
  int c;

  c=0;
  while (n) {
    c++;
    n&=(n-1);
  }
  return(c);
}



void ClampOutputVolume(int *v) {

  if (*v>ECHOGAIN_MAXOUT)
    *v=ECHOGAIN_MAXOUT;
  else if (*v<ECHOGAIN_MINOUT)
    *v=ECHOGAIN_MINOUT;
}



// -128 dB means muted
int Add_dB (int a, int b) {

  if (a==ECHOGAIN_MINOUT || b==ECHOGAIN_MINOUT)
    return(ECHOGAIN_MINOUT);
  return(a+b);
}



int ADATmode() {
  return(!memcmp(dmodeName[dmodeVal], "ADAT", 4));
}



// Write an enumerated ALSA control
int SetEnum(int numid, int val) {
  int err;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, numid);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_enumerated(control, 0, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
  return(err);
}



// Read an enumerated ALSA control
int GetEnum(int numid) {
  int err, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, numid);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control)) < 0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  val=snd_ctl_elem_value_get_enumerated(control, 0);
  return(val);
}



// Turn VU-meters on/off
void SetVUmeters(int onoff) {
  static signed char oncount=0;
  int err;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (onoff)
    oncount++;
  else
    if (--oncount<0)
      oncount=0;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, vuswitchId);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, !!oncount);
  if ((err=snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}



void GetVUmeters(int *InLevel, int *InPeak, int *OutLevel, int *OutPeak) {
  int err, i, m;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_id_set_numid(id, vumetersId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err = snd_ctl_elem_read(ctlhandle, control)) < 0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
    return;
  }

  m=0;
  for (i=0; i<nLOut; i++) {
    OutLevel[i]=snd_ctl_elem_value_get_integer(control, m++);
    OutPeak[i]=snd_ctl_elem_value_get_integer(control, m++);
  }

  m=1*metersNumber*metersTypes;
  for (i=0; i<nIn; i++) {
    InLevel[i]=snd_ctl_elem_value_get_integer(control, m++);
    InPeak[i]=snd_ctl_elem_value_get_integer(control, m++);
  }
/*
  if (metersStreams==3) {       // Has PCM levels (Mia only) ?
    m=2*metersNumber*metersTypes;
    for (i=0; i<nPOut; i++) {
      PcmLevel[i]=snd_ctl_elem_value_get_integer(control, m++);
      PcmPeak[i]=snd_ctl_elem_value_get_integer(control, m++);
    }
  }
*/
}



#ifdef REVERSE

// Enable/disable widgets that control ADAT digital channels
void SetSensitivity(int enable) {
  int i;

  for (i=fdOut+2; i<nLOut; i++) {
    if (!enable && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mixout[i])))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixout[0]), TRUE);
    gtk_widget_set_sensitive(mixout[i], enable);
    gtk_widget_set_sensitive(outvol_lbl[i], enable);
    gtk_widget_set_sensitive(outvol[i], enable);
    gtk_widget_set_sensitive(mstvol_lbl[i], enable);
    gtk_widget_set_sensitive(mstvol[i], enable);
  }
  for (i=fdIn+2; i<nIn; i++) {
    gtk_widget_set_sensitive(mixvol_lbl[i], enable);
    gtk_widget_set_sensitive(mixvol[i], enable);
  }
  if (!enable && InputChannel>=fdIn+2)
    InputChannel=0;
}

#else

// Enable/disable widgets that control ADAT digital channels
void SetSensitivity(int enable) {
  int i;

  for (i=fdIn+2; i<nIn; i++) {
    if (!enable && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mixin[i])))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixin[0]), TRUE);
    gtk_widget_set_sensitive(mixin[i], enable);
  }
  for (i=fdOut+2; i<nLOut; i++) {
    gtk_widget_set_sensitive(mixvol_lbl[i], enable);
    gtk_widget_set_sensitive(mixvol[i], enable);
    gtk_widget_set_sensitive(outvol_lbl[i], enable);
    gtk_widget_set_sensitive(outvol[i], enable);
    gtk_widget_set_sensitive(mstvol_lbl[i], enable);
    gtk_widget_set_sensitive(mstvol[i], enable);
  }
  if (!enable && OutputChannel>=fdOut+2)
    OutputChannel=0;
}

#endif


// At startup this functions reads the current volume levels and sets the widgets accordingly.
void InitPCMVolumeGUI(int inout, char load) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, ch, channels, volId, val, fix, *vol;
  char str[16], *fmt;
  GtkObject **adj;
  GtkWidget **label;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);

  if (inout==INPUT) {
    channels=fdIn;
    label=invol_lbl;
    fmt="%+4.1f";
    adj=involadj;
    volId=involId;
    fix=ECHOGAIN_MININP+ECHOGAIN_MAXINP;
    vol=LineInVol;
  } else {
    channels=nPOut;
    label=outvol_lbl;
    fmt="%+3.0f";
    adj=outvoladj;
    volId=outvolId;
    fix=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT;
    vol=PcmOutVol;
  }

  snd_ctl_elem_id_set_numid(id, volId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
    return;
  }

  if (!load) {
    for (ch=0; ch<channels; ch++)
      vol[ch]=snd_ctl_elem_value_get_integer(control, ch);
  }

  for (ch=0; ch<channels; ch++) {
    val=vol[ch];
    if (val==ECHOGAIN_MINOUT)
      strcpy(str, "mute");
    else
      sprintf(str, fmt, (float)val/256.0);
    gtk_label_set_text(GTK_LABEL(label[ch]), str);
    val=fix-val;     // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(adj[ch]), (gfloat)val);
  }
/*
  for (ch=0; ch<channels; ch++) {
    val=snd_ctl_elem_value_get_integer(control, ch);
if (load&&inout==INPUT)  val=LineInVol[ch];
if (load&&inout==OUTPUT)  val=PcmOutVol[ch];
    if (val==ECHOGAIN_MINOUT)
      strcpy(str, "mute");
    else
      sprintf(str, fmt, (float)val/256.0);
    gtk_label_set_text(GTK_LABEL(label[ch]), str);
    val=fix-val;     // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(adj[ch]), (gfloat)val);
  }*/
}



void InitLineVolumeGUI() {
  int ch;

  for (ch=0; ch<nLOut; ch++) {
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mstadj[ch]), (gfloat)(ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-LineOutVol[ch]));
  }
}



// At startup this functions reads the current nominal levels and sets the switches accordingly.
void InitNominalLevelGUI(int numid) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, i, n;
  GtkWidget **w;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  if (numid==p4InId) {
    snd_ctl_elem_id_set_numid(id, p4InId);
    n=fdIn;
    w=p4dbuIn;
  } else {
    snd_ctl_elem_id_set_numid(id, p4OutId);
    n=fdOut;
    w=p4dbuOut;
  }
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  for (i=0; i<n; i++)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w[i]), !snd_ctl_elem_value_get_integer(control, i));     // FALSE is +4 here
}



// At startup this functions reads if the dithering is enabled sets the button accordingly.
void InitDitheringGUI(int numid) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, ditherId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ditherToggle), snd_ctl_elem_value_get_integer(control, 0));
}



// Read current mixer settings.
void InitMixer(void) {
  int err, in, out;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);

  for (out=0; out<nLOut; out++) {
    for (in=0; in<nIn; in++) {
      snd_ctl_elem_id_set_numid(id, Mixer[out][in].id);
      snd_ctl_elem_value_set_id(control, id);
      if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
        printf("Control %s element read error: %s\n", card, snd_strerror(err));
      Mixer[out][in].val=snd_ctl_elem_value_get_integer(control, 0);
    }
  }
}



#define XCELLBORDER 2
#define YCELLBORDER 2
#define XCELLDIM 20
#define YCELLDIM 32
#define XCELLTOT (1+XCELLBORDER*2+XCELLDIM)
#define YCELLTOT (1+YCELLBORDER*2+YCELLDIM)
#define XVOLUME 6
#define XMETER 14

// Draw the matrix mixer
gint DrawMixer(gpointer data) {
  GdkRectangle update_rect;
  int InLevel[ECHO_MAXAUDIOINPUTS];
  int InPeak[ECHO_MAXAUDIOINPUTS];
  int OutLevel[ECHO_MAXAUDIOOUTPUTS];
  int OutPeak[ECHO_MAXAUDIOOUTPUTS];
  static int InClip[ECHO_MAXAUDIOINPUTS];
  static int OutClip[ECHO_MAXAUDIOOUTPUTS];
  char str[8];
  int i, o, x, dB, db, inchannels, outchannels;
  GdkColor Grid={0x9694C4, 0, 0, 0};
  GdkColor Bars={0x00FF00, 0, 0, 0};
  GdkColor Bars1={0x000000, 0, 0, 0};
  GdkColor Peak={0x1BABFF, 0, 0, 0};
  GdkColor Level={0xC0B000, 0, 0, 0};
  GdkColor Hilight={0x000078, 0, 0, 0};

#define MINXI 30
#define MINXO 132
#define MINY 20

  if (ADATmode()) {
    inchannels=nIn;
    outchannels=nLOut;
  } else {
    inchannels=fdIn+2;
    outchannels=fdOut+2;
  }

  if (!Mixpixmap)
    return(TRUE);

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = Mixwidth;
  update_rect.height = Mixheight;
  GetVUmeters(InLevel, InPeak, OutLevel, OutPeak);

  if (!gc) {
    gc=gdk_gc_new(gtk_widget_get_parent_window(Mixdarea));
    for (i=0; i<nIn; i++)
      InClip[i]=0;
    for (i=0; i<nLOut; i++)
      OutClip[i]=0;
  }

  gdk_draw_rectangle(Mixpixmap, Mixdarea->style->black_gc, TRUE, 0, 0, Mixwidth, Mixheight);

/*
  gdk_draw_rectangle(Mixpixmap, Mixdarea->style->black_gc, TRUE, 0, 0, Mixwidth, Mixheight);

  gdk_gc_set_foreground(gc, &Grid);
  gdk_gc_set_font(gc, fnt);
  gdk_draw_string(Mixpixmap, fnt, gc, 2, MINY-12+4, "  dB");
  for (i=0; i<=120; i+=12) {
    sprintf(str, "%4d", -i);
    gdk_draw_string(Mixpixmap, fnt, gc, 2, MINY+i+4, str);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, MINXI, MINY+i, Grwidth-MINXI, 1);
  }
  gdk_draw_rectangle(Mixpixmap, gc, TRUE, MINXI, MINY+128, Grwidth-MINXI, 1);
*/
  x=0;

  // Highlight
  gdk_gc_set_foreground(gc, &Hilight);
  gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*InputChannel, XCELLTOT*(OutputChannel+1), YCELLTOT);
  gdk_draw_rectangle(Mixpixmap, gc, TRUE, XCELLTOT*(OutputChannel+1), YCELLTOT*InputChannel, XCELLTOT, Mixheight);

  // Draw the grid
  gdk_gc_set_font(gc, fnt);
  gdk_gc_set_foreground(gc, &Grid);
  for (i=0; i<nIn; i++) {
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*(i+1)-1, Mixwidth, 1);
    if (i<fdIn)
      sprintf(str, "A%d", i);
    else
      sprintf(str, "D%d", i-fdIn);
    gdk_draw_string(Mixpixmap, fnt, gc, 1, YCELLTOT*i+(YCELLTOT>>1)+4, str);
  }
  for (o=0; o<nLOut; o++) {
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XCELLTOT*(o+1), 0, 1, Mixheight);
    if (o<fdOut)
      sprintf(str, "A%d", o);
    else
      sprintf(str, "D%d", o-fdOut);
    gdk_draw_string(Mixpixmap, fnt, gc, XCELLTOT*(o+1)+(XCELLTOT/2)-6, YCELLTOT*GMixerSection.Line+YCELLTOT+8, str);
  }
  gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*(GMixerSection.Line+1)-1, Mixwidth, 1);
  gdk_draw_string(Mixpixmap, fnt, gc, 1, 8, "In");
  gdk_draw_string(Mixpixmap, fnt, gc, 1, YCELLTOT*GMixerSection.Line+YCELLTOT+8, "Out");

  // Draw input levels and peaks
  for (i=0; i<inchannels; i++) {
    dB=InLevel[i]>>8;
    db=dB>>2;
    gdk_gc_set_foreground(gc, &Bars);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
    if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
      gdk_gc_set_foreground(gc, &Bars1);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db-1, BARWIDTH, 1);
    }
  }
  gdk_gc_set_foreground(gc, &Peak);
  for (i=0; i<inchannels; i++) {
    db=InPeak[i]>>10;
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db, BARWIDTH, 1);
  }

  // Draw output levels, peaks and volumes
  for (o=0; o<outchannels; o++) {
    dB=OutLevel[o]>>8;
    db=dB>>2;
    gdk_gc_set_foreground(gc, &Bars);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*GMixerSection.Line+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
    if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
      gdk_gc_set_foreground(gc, &Bars1);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*GMixerSection.Line+YCELLBORDER-db-1, BARWIDTH, 1);
    }
    db=OutPeak[o]>>10;
    gdk_gc_set_foreground(gc, &Peak);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*GMixerSection.Line+YCELLBORDER-db, BARWIDTH, 1);
    gdk_gc_set_foreground(gc, &Level);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME+XCELLTOT*(o+1), YCELLTOT*GMixerSection.Line+YCELLBORDER, 1, YCELLDIM);
    db=LineOutVol[o]>>10;
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME-2+XCELLTOT*(o+1), YCELLTOT*GMixerSection.Line+YCELLBORDER-db, 5, 1);
  }

  // Draw monitor mixer elements
  for (o=0; o<outchannels; o++) {
    for (i=0; i<inchannels; i++) {
      dB=Add_dB(Mixer[o][i].val, InLevel[i])>>8;
      db=dB>>2;
      if (db<-YCELLDIM)
        db=-YCELLDIM;
      gdk_gc_set_foreground(gc, &Bars);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
      if (dB>-(YCELLDIM<<2))
        if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
          gdk_gc_set_foreground(gc, &Bars1);
          gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db-1, BARWIDTH, 1);
        }
      gdk_gc_set_foreground(gc, &Level);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER, 1, YCELLDIM);
      db=Mixer[o][i].val>>10;
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME-2+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db, 5, 1);
    }
  }
  
  gtk_widget_draw(Mixdarea, &update_rect);
  return(TRUE);
}



// Draw the VU-meter
gint DrawVUmeters(gpointer data) {
  GdkRectangle update_rect;
  int InLevel[ECHO_MAXAUDIOINPUTS];
  int InPeak[ECHO_MAXAUDIOINPUTS];
  int OutLevel[ECHO_MAXAUDIOOUTPUTS];
  int OutPeak[ECHO_MAXAUDIOOUTPUTS];
  static int InClip[ECHO_MAXAUDIOINPUTS];
  static int OutClip[ECHO_MAXAUDIOOUTPUTS];
  int i, x, dB;
  char str[16];
  GdkColor Selected={0xC86060, 0, 0, 0};
  GdkColor Grid={0x9694C4, 0, 0, 0};
  GdkColor Grid2={0x646383, 0, 0, 0};
  GdkColor dBValues={0x00B000, 0, 0, 0};
  GdkColor AnBars={0x00E0B8, 0, 0, 0};
  GdkColor DiBars={0x98E000, 0, 0, 0};
  GdkColor ClipPeak={0, 0, 0, 0};
  GdkColor Peak={0x00FF00, 0, 0, 0};

#define MINXI 30
#define MINXO 132
#define MINY 20

  if (!VUpixmap)
    return(TRUE);

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = VUwidth;
  update_rect.height = VUheight;
  GetVUmeters(InLevel, InPeak, OutLevel, OutPeak);

  if (!gc) {
    gc=gdk_gc_new(gtk_widget_get_parent_window(VUdarea));
    for (i=0; i<nIn; i++)
      InClip[i]=0;
    for (i=0; i<nLOut; i++)
      OutClip[i]=0;
  }

  // Clear the image
  gdk_draw_rectangle(VUpixmap, VUdarea->style->black_gc, TRUE, 0, 0, VUwidth, VUheight);

  // Draw the dB scale and the grid
  gdk_gc_set_font(gc, fnt);
  gdk_gc_set_foreground(gc, &Peak);
  gdk_draw_string(VUpixmap, fnt, gc, 2, MINY-12+4, "  dB");
  for (i=0; i<=120; i+=12) {
    sprintf(str, "%4d", -i);
    gdk_gc_set_foreground(gc, &dBValues);
    gdk_draw_string(VUpixmap, fnt, gc, 2, MINY+i+4, str);
    gdk_gc_set_foreground(gc, &Grid);
    gdk_draw_rectangle(VUpixmap, gc, TRUE, MINXI, MINY+i, VUwidth-MINXI, 1);
  }
  gdk_gc_set_foreground(gc, &Grid2);
  gdk_draw_rectangle(VUpixmap, gc, TRUE, MINXI, MINY+128, VUwidth-MINXI, 1);

  x=MINXI+BARSEP;

  for (i=0; i<nIn; i++) {
    if (i<fdIn)
      gdk_gc_set_foreground(gc, &AnBars);
    else
      gdk_gc_set_foreground(gc, &DiBars);
    dB=InLevel[i]>>8;
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 129+MINY-(MINY-dB));

    dB=InPeak[i]>>8;
    if (dB==0)
      InClip[i]=64;
    if (InClip[i]) {
      InClip[i]--;
      ClipPeak.pixel=(InClip[i]<<18)+((255-(InClip[i]*3))<<8);
      gdk_gc_set_foreground(gc, &ClipPeak);
    } else {
      gdk_gc_set_foreground(gc, &Peak);
    }
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 1);
    if (InputChannel==i) {
      gdk_gc_set_foreground(gc, &Selected);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x+1, MINY+128+3, BARWIDTH-2, 1);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY+128+4, BARWIDTH, 1);
    }
    x+=BARWIDTH+BARSEP;
  }

  x+=BARWIDTH+BARSEP;
  for (i=0; i<nLOut; i++) {
    if (i<fdOut)
      gdk_gc_set_foreground(gc, &AnBars);
    else
      gdk_gc_set_foreground(gc, &DiBars);
    dB=OutLevel[i]>>8;
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 129+MINY-(MINY-dB));

    dB=OutPeak[i]>>8;
    if (dB==0)
      OutClip[i]=64;
    if (OutClip[i]) {
      OutClip[i]--;
      ClipPeak.pixel=(OutClip[i]<<18)+((255-(OutClip[i]*3))<<8);
      gdk_gc_set_foreground(gc, &ClipPeak);
    } else {
      gdk_gc_set_foreground(gc, &Peak);
    }
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 1);
    if (OutputChannel==i) {
      gdk_gc_set_foreground(gc, &Selected);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x+1, MINY+128+3, BARWIDTH-2, 1);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY+128+4, BARWIDTH, 1);
    }
    x+=BARWIDTH+BARSEP;
  }
  gtk_widget_draw(VUdarea, &update_rect);

  return(TRUE);
}



///////////////////// GUI events


void Input_selector_clicked(GtkWidget *widget, gpointer ich) {
  int och, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (InputChannel==(int)ich)
    return;

  InputChannel=(int)ich;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  for (och=0; och<nLOut; och++) {
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-Mixer[och][InputChannel].val;   // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[och]), (gfloat)val);
  }
}



void Output_selector_clicked(GtkWidget *widget, gpointer och) {
  int ich, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (OutputChannel==(int)och)
    return;

  OutputChannel=(int)och;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  for (ich=0; ich<nIn; ich++) {
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-Mixer[OutputChannel][ich].val;  // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[ich]), (gfloat)val);
  }
}


#ifdef REVERSE

static gint Mixer_button_press(GtkWidget *widget, GdkEventButton *event) {
  int ch;

  ch=event->x/XCELLTOT-1;
  if (ch!=OutputChannel) {
    if (ch<nLOut && (ADATmode() || ch<fdOut+2))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixout[ch]), TRUE);
  }
  GMixerRow=event->y/YCELLTOT;
  if (GMixerRow!=InputChannel) {
    if (GMixerRow<=GMixerSection.Monitor)
      gtk_widget_grab_focus(GTK_WIDGET(mixvol[GMixerRow]));
  }

  if (event->button==1) {
    mouseY=event->y;
    mouseButton=1;
  }
  return(TRUE);
}

#else //REVERSE

static gint Mixer_button_press(GtkWidget *widget, GdkEventButton *event) {
  int ch;

  GMixerRow=event->y/YCELLTOT;
  if (GMixerRow!=InputChannel) {
    if (GMixerRow<=GMixerSection.Monitor)
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixin[GMixerRow]), TRUE);
/*    else if (GMixerRow==GMixerSection.Line)
      gtk_widget_grab_focus(GTK_WIDGET(mstvol[GMixerRow]));*/
  }
  ch=event->x/XCELLTOT-1;
  if (ch!=OutputChannel) {
    if (ch<nLOut && (ADATmode() || ch<fdOut+2))
      gtk_widget_grab_focus(GTK_WIDGET(mixvol[ch]));
  }

  if (event->button==1) {
    mouseY=event->y;
    mouseButton=1;
  }
  return(TRUE);
}

#endif //REVERSE



static gint Mixer_button_release(GtkWidget *widget, GdkEventButton *event) {

  if (event->state & GDK_BUTTON1_MASK)
    mouseButton=0;
  return TRUE;
}



static gint Mixer_motion_notify(GtkWidget *widget, GdkEventMotion *event) {
  int x, y;
  GdkModifierType state;
  float val;

  if (event->is_hint)
    gdk_window_get_pointer(event->window, &x, &y, &state);
  else {
    x=event->x;
    y=event->y;
    state=event->state;
  }
  
  if (GMixerRow<=GMixerSection.Monitor) {
    if (mouseButton && Mixpixmap != NULL) {
      val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-Mixer[OutputChannel][InputChannel].val;       // Invert
      val+=256*(y-mouseY);
      mouseY=y;
      // Gtk already limits the range of "val"
#ifdef REVERSE
      gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[InputChannel]), (gfloat)val);
#else
      gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[OutputChannel]), (gfloat)val);
#endif
    }
  } else if (GMixerRow==GMixerSection.Line) {
    if (mouseButton && Mixpixmap != NULL) {
      val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-LineOutVol[OutputChannel];                    // Invert
      val+=256*(y-mouseY);
      mouseY=y;
      // Gtk already limits the range of "val"
      gtk_adjustment_set_value(GTK_ADJUSTMENT(mstadj[OutputChannel]), (gfloat)val);
    }
  }

  return(TRUE);
}



void Monitor_volume_changed(GtkWidget *widget, gpointer cnl) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  //unsigned int count;
  int err, val, rval, ch;
  char str[16];
  struct mixel *mxl;
  //snd_ctl_elem_type_t type;
  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

//printf("Monitor_volume_changed()  %d %d\n",InputChannel,OutputChannel);
  val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value

  ch=(int)cnl;
  if (ch<ECHO_MAXAUDIOINPUTS) {
    mxl=&Mixer[OutputChannel][ch];
    rval=Add_dB(val, LineOutVol[OutputChannel]);
  } else {
    ch-=ECHO_MAXAUDIOINPUTS;
    mxl=&Mixer[ch][InputChannel];
    rval=Add_dB(val, LineOutVol[ch]);
  }

  ClampOutputVolume(&rval);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_id_set_numid(id, mxl->id);
/*
  snd_ctl_elem_info_set_id(info, id);
  if ((err = snd_ctl_elem_info(handle, info)) < 0) {
    printf("Control %s cinfo error: %s\n", card, snd_strerror(err));
  }
  snd_ctl_elem_info_get_id(info, id);
  type = snd_ctl_elem_info_get_type(info);
  count = snd_ctl_elem_info_get_count(info);
//printf("%s\n", snd_ctl_elem_info_get_name(info));
*/
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, rval);
  if ((err = snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  } else {
    mxl->val=val;
    if (val==ECHOGAIN_MINOUT)
      strcpy(str, "mute");
    else
      sprintf(str, "%+3.0f", (float)val/256.0);
    gtk_label_set_text(GTK_LABEL(mixvol_lbl[ch]), str);
  }
}



void Monitor_outvolume_clicked(GtkWidget *widget, gpointer ch) {

  OutputChannel=(int)ch;
}



void Monitor_involume_clicked(GtkWidget *widget, gpointer ch) {

  InputChannel=(int)ch;
}



void Gang_button_toggled(GtkWidget *widget, gpointer x) {

  Gang=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}



void PCM_volume_changed(GtkWidget *widget, gpointer ch) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  char str[16], *fmt;
  int err, channel, val, rval;
  GtkWidget **label;
  GtkObject **adj;

  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  if ((int)ch<ECHO_MAXAUDIOINPUTS) {
    // Input
    channel=(int)ch;
    label=invol_lbl;
    fmt="%+4.1f";
    adj=involadj;
    rval=val=ECHOGAIN_MININP+ECHOGAIN_MAXINP-(int)GTK_ADJUSTMENT(widget)->value;        // Inverts the value
    LineInVol[channel]=val;
    snd_ctl_elem_id_set_numid(id, involId);
  } else {
    // Output
    channel=(int)ch-ECHO_MAXAUDIOINPUTS;
    label=outvol_lbl;
    fmt="%+3.0f";
    adj=outvoladj;
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;             // Inverts the value
    PcmOutVol[channel]=val;
    rval=Add_dB(val, LineOutVol[channel]);
    snd_ctl_elem_id_set_numid(id, outvolId);
  }

  ClampOutputVolume(&rval);

  if (val==ECHOGAIN_MINOUT)
    strcpy(str, "mute");
  else
    sprintf(str, fmt, (float)val/256.0);
  gtk_label_set_text(GTK_LABEL(label[channel]), str);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_value_set_id(control, id);

  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  snd_ctl_elem_value_set_integer(control, channel, rval);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0)
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  if (Gang)
    gtk_adjustment_set_value(GTK_ADJUSTMENT(adj[channel^0x01]), (gfloat)GTK_ADJUSTMENT(widget)->value);
}



// Changes the PCM volume according to the current Line-out volume for non-vmixer cards
void UpdatePCMVolume(int outchannel) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val;

  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_numid(id, outvolId);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_value_set_id(control, id);

  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));

  val=Add_dB(PcmOutVol[outchannel], LineOutVol[outchannel]);
  ClampOutputVolume(&val);

  snd_ctl_elem_value_set_integer(control, outchannel, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0)
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
}



// Changes the monitor mixer volume according to the current Line-out volume for non-vmixer cards
void UpdateMixerVolume(int outchannel) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val, ch;
  
  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);

  for (ch=0; ch<nIn; ch++) {
    val=Add_dB(Mixer[outchannel][ch].val, LineOutVol[outchannel]);
    ClampOutputVolume(&val);
    snd_ctl_elem_id_set_numid(id, Mixer[outchannel][ch].id);
    snd_ctl_elem_value_set_id(control, id);
    snd_ctl_elem_value_set_integer(control, 0, val);
    if ((err = snd_ctl_elem_write(ctlhandle, control)) < 0)
      printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}



void Line_volume_changed(GtkWidget *widget, gpointer ch) {
  char str[16];
  int channel, val;

  channel=(int)ch;

  val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value
  LineOutVol[channel]=val;

  if (val==ECHOGAIN_MINOUT)
    strcpy(str, "mute");
  else
    sprintf(str, "%+.0f", (float)val/256.0);
  gtk_label_set_text(GTK_LABEL(mstvol_lbl[channel]), str);

  if (0) {//has_vmixer
  } else {
    UpdatePCMVolume(channel);
    UpdateMixerVolume(channel);
  }

  if (Gang)
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mstadj[channel^0x01]), (gfloat)GTK_ADJUSTMENT(widget)->value);
}


/*
void Vmixer_volume_changed(GtkWidget *widget, gpointer idx) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  char str[16];
  int err, val, pipe, output, index;

  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  index=(int)idx;
  pipe=index>>1;
  output=index&1;

  val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;       // Inverts the value

  if (val==ECHOGAIN_MINOUT)
    strcpy(str, "mute");
  else
    sprintf(str, "%+.0f", (float)val/256.0);
  gtk_label_set_text(GTK_LABEL(vmixer_lbl[index]), str);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_value_set_id(control, id);

  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  snd_ctl_elem_value_set_integer(control, pipe*nPOut+output, val);
//  if (Gang)
//    snd_ctl_elem_value_set_integer(control, (pipe*nPOut+output)^0x01, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0)
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  if (Gang)
    gtk_adjustment_set_value(GTK_ADJUSTMENT(vmixeradj[index^0x01]), (gfloat)GTK_ADJUSTMENT(widget)->value);
}
*/


void Nominal_level_toggled(GtkWidget *widget, gpointer ch) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val, item;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  val=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  if ((int)ch<ECHO_MAXAUDIOINPUTS) {
    item=(int)ch;
    snd_ctl_elem_id_set_numid(id, p4InId);
  } else {
    item=(int)ch-ECHO_MAXAUDIOINPUTS;
    snd_ctl_elem_id_set_numid(id, p4OutId);
  }
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0)
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  snd_ctl_elem_value_set_integer(control, item, !val);  // FALSE is +4
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0)
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
}



void Dithering_toggled(GtkWidget *widget, gpointer g) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  val=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, ditherId);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}



void Digital_mode_activate(GtkWidget *widget, gpointer mode) {
  int adat;
  
  if (SetEnum(dmodeId, (int)mode)<0)
    // Restore old value if it failed
    gtk_option_menu_set_history(GTK_OPTION_MENU(dmodeOpt), dmodeVal);
  else {
    dmodeVal=(int)mode;
    // When I change the digital mode, the clock source can change too
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal=GetEnum(clocksrcId));
  }
  adat=ADATmode();
  SetSensitivity(adat);
  if (adat)
    GMixerSection.Monitor=nIn-1;
  else
    GMixerSection.Monitor=fdIn+2-1;     // S/PDIF has only 2 channels
}



void Clock_source_activate(GtkWidget *widget, gpointer clk) {

  if (SetEnum(clocksrcId, (int)clk)<0)
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal);
  else
    clocksrcVal=(int)clk;
}



void SPDIF_mode_activate(GtkWidget *widget, gpointer mode) {

  SetEnum(spdifmodeId, (int)mode);      // This one should never fail
  spdifmodeVal=(int)mode;
}



void GetChannels(void) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_HWDEP);
  snd_ctl_elem_id_set_numid(id, channelsId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err = snd_ctl_elem_read(ctlhandle, control)) < 0) {
    printf("GetChannels() read error: %s\n", snd_strerror(err));
    exit(1);
  }
  nIn=snd_ctl_elem_value_get_integer(control, 0);       // Number of input channels
  fdIn=snd_ctl_elem_value_get_integer(control, 1);      // First digital in (= number of analog input channels)
  nLOut=snd_ctl_elem_value_get_integer(control, 2);     // Number of output channels
  fdOut=snd_ctl_elem_value_get_integer(control, 3);     // First digital out
  nPOut=snd_ctl_elem_value_get_integer(control, 4);     // Number of virtual output channels (==nLOut on non-vmixer cards)
}



// Create a new backing pixmap of the appropriate size
static gint VU_configure_event(GtkWidget *widget, GdkEventConfigure *event) {

  if (VUpixmap)
    gdk_pixmap_unref(VUpixmap);
  VUpixmap=gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle(VUpixmap, widget->style->black_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  return(TRUE);
}



// Redraw the screen from the backing pixmap
static gint VU_expose(GtkWidget *widget, GdkEventExpose *event) {

  if (VUpixmap)
    gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], VUpixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);
  return(FALSE);
}



// Create a new backing pixmap of the appropriate size
static gint Mixer_configure_event(GtkWidget *widget, GdkEventConfigure *event) {
  
  if (Mixpixmap)
    gdk_pixmap_unref(Mixpixmap);
  Mixpixmap=gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle(Mixpixmap, widget->style->black_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  return(TRUE);
}



// Redraw the screen from the backing pixmap
static gint Mixer_expose(GtkWidget *widget, GdkEventExpose *event) {
  
  if (Mixpixmap)
    gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], Mixpixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);
  return(FALSE);
}



gint window_close(GtkWidget *widget, GdkEvent *event, gpointer geom) {
  struct geometry *g=geom;

  gdk_window_get_root_origin(widget->window, &g->x, &g->y);
  gdk_window_get_size(widget->window, &g->w, &g->h);
  gtk_widget_hide(widget);
  return(TRUE);
}



gint Mainwindow_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {

  if (VUwindow) {
    gdk_window_get_root_origin(VUwindow->window, &VUw_geom.x, &VUw_geom.y);
    gtk_widget_destroy(VUwindow);
  }
  if (GMwindow) {
    gdk_window_get_root_origin(GMwindow->window, &GMw_geom.x, &GMw_geom.y);
    gtk_widget_destroy(GMwindow);
  }
  gtk_main_quit();
  return(FALSE);
}



gint VUwindow_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {

  gdk_window_get_root_origin(VUwindow->window, &VUw_geom.x, &VUw_geom.y);
  VUw_geom.st=0;
  return(FALSE);
}



gint GMwindow_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {

  gdk_window_get_root_origin(GMwindow->window, &GMw_geom.x, &GMw_geom.y);
  GMw_geom.st=0;
  return(FALSE);
}



gint VUwindow_destroy(GtkWidget *widget, gpointer data) {

  SetVUmeters(0);
  gtk_timeout_remove(VUtimer);
  //@@del gc and fnt
  VUwindow=0;
  return(TRUE);
}



gint GMwindow_destroy(GtkWidget *widget, gpointer data) {

  SetVUmeters(0);
  gtk_timeout_remove(Mixtimer);
  //@@del gc and fnt
  GMwindow=0;
  return(TRUE);
}



void VUmeters_button_click(GtkWidget *widget, gpointer n) {

  if (!VUwindow) {
    // Create VU-meter window
    VUwidth=MINXI+(BARWIDTH+BARSEP)*(nIn+nLOut+1)+BARSEP;
    VUheight=160;
    SetVUmeters(1);
    VUwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (VUwindow), "VU-meters");
    gtk_signal_connect(GTK_OBJECT(VUwindow), "destroy", GTK_SIGNAL_FUNC(VUwindow_destroy), NULL);
    gtk_signal_connect(GTK_OBJECT(VUwindow), "delete_event", GTK_SIGNAL_FUNC(VUwindow_delete), NULL);
    gtk_window_set_policy(GTK_WINDOW(VUwindow), FALSE, FALSE, TRUE);
    if (VUw_geom.st!=NOPOS)
      gtk_widget_set_uposition(VUwindow, VUw_geom.x, VUw_geom.y);
    gtk_widget_show(VUwindow);

    VUdarea=gtk_drawing_area_new();
    gtk_widget_set_events(VUdarea, GDK_EXPOSURE_MASK);
    gtk_drawing_area_size(GTK_DRAWING_AREA(VUdarea), VUwidth, VUheight);
    gtk_container_add(GTK_CONTAINER(VUwindow), VUdarea);

    gtk_widget_show(VUdarea);
    gtk_signal_connect(GTK_OBJECT(VUdarea), "expose_event", (GtkSignalFunc)VU_expose, NULL);
    gtk_signal_connect(GTK_OBJECT(VUdarea), "configure_event", (GtkSignalFunc)VU_configure_event, NULL);
    VUtimer=gtk_timeout_add(30, DrawVUmeters, 0);       // The hw updates the meters about 30 times/s
    gdk_window_clear_area(VUdarea->window, 0, 0, VUwidth, VUheight);
    VUw_geom.st=1;
  }
}



void GMixer_button_click(GtkWidget *widget, gpointer n) {

  if (!GMwindow) {
    // Create VU-meter window
    Mixwidth=XCELLTOT*(nLOut+1);
    Mixheight=YCELLTOT*(GMixerSection.Line+1)+9;
    SetVUmeters(1);
    GMwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (GMwindow), "Mixer");
    gtk_signal_connect(GTK_OBJECT(GMwindow), "destroy", GTK_SIGNAL_FUNC(GMwindow_destroy), NULL);
    gtk_signal_connect(GTK_OBJECT(GMwindow), "delete_event", GTK_SIGNAL_FUNC(GMwindow_delete), NULL);
    gtk_window_set_policy(GTK_WINDOW(GMwindow), FALSE, FALSE, TRUE);
    if (GMw_geom.st!=NOPOS)
      gtk_widget_set_uposition(GMwindow, GMw_geom.x, GMw_geom.y);
    gtk_widget_show(GMwindow);

    Mixdarea=gtk_drawing_area_new();
    gtk_widget_set_events(Mixdarea, GDK_EXPOSURE_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
    gtk_drawing_area_size(GTK_DRAWING_AREA(Mixdarea), Mixwidth, Mixheight);
    gtk_container_add(GTK_CONTAINER(GMwindow), Mixdarea);

    gtk_widget_show(Mixdarea);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "expose_event", (GtkSignalFunc)Mixer_expose, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "configure_event", (GtkSignalFunc)Mixer_configure_event, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "motion_notify_event", (GtkSignalFunc)Mixer_motion_notify, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "button_press_event", (GtkSignalFunc)Mixer_button_press, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "button_release_event", (GtkSignalFunc)Mixer_button_release, NULL);
    Mixtimer=gtk_timeout_add(30, DrawMixer, 0);         // The hw updates the meters about 30 times/s
    gdk_window_clear_area(Mixdarea->window, 0, 0, Mixwidth, Mixheight);
    GMw_geom.st=1;
  }
}



void Mixer_button_click(GtkWidget *widget, gpointer n) {

  gtk_widget_show(Mixerwindow);
}



void PV_button_click(GtkWidget *widget, gpointer n) {

  gtk_widget_show(PVwindow);
}



void LV_button_click(GtkWidget *widget, gpointer n) {

  gtk_widget_show(LVwindow);
}



// Scan all controls and store their ID's for later use.
int OpenControls(const char *card, const char *cardname) {
  int err, i, o;
  int FirstMixerId, FirstVmixerId, numid, count, items, item;
  snd_hctl_t *handle;
  snd_hctl_elem_t *elem;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_info_t *info;

  outvolId=masterId=vmixerId=p4InId=p4OutId=dmodeId=clocksrcId=spdifmodeId=vuswitchId=vumetersId=0;
  outvolCount=0;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_info_alloca(&info);

  ndmodes=nclocksrc=nspdifmodes=0;

  FirstMixerId=FirstVmixerId=-1;
  if ((err=snd_hctl_open(&handle, card, 0))<0) {
    printf("Control %s open error: %s", card, snd_strerror(err));
    return err;
  }
  if ((err=snd_hctl_load(handle))<0) {
    printf("Control %s local error: %s\n", card, snd_strerror(err));
    return err;
  }
  for (elem=snd_hctl_first_elem(handle); elem; elem=snd_hctl_elem_next(elem)) {
    if ((err=snd_hctl_elem_info(elem, info))<0) {
      printf("Control %s snd_hctl_elem_info error: %s\n", card, snd_strerror(err));
      return err;
    }
    if (snd_ctl_elem_info_is_inactive(info))
      continue;
    snd_hctl_elem_get_id(elem, id);
    numid=snd_ctl_elem_id_get_numid(id);
    count=snd_ctl_elem_info_get_count(info);
    if (!strcmp("Monitor Mixer Volume", snd_ctl_elem_id_get_name(id))) {
      if (FirstMixerId<0) {
        FirstMixerId=numid;
printf("First Mixer id=%d\n", FirstMixerId);
        mixerOutputs=snd_ctl_elem_info_get_dimension(info, 0);
        mixerInputs=snd_ctl_elem_info_get_dimension(info, 1);
      }
    } else if (!strcmp("VMixer Volume", snd_ctl_elem_id_get_name(id))) {
      if (FirstVmixerId<0) {
        FirstVmixerId=numid;
printf("First Vmixer id=%d\n", FirstVmixerId);
        vmixerPipes=snd_ctl_elem_info_get_dimension(info, 0);
        vmixerOutputs=snd_ctl_elem_info_get_dimension(info, 1);
      }
    } else if (!strcmp("PCM Playback Volume", snd_ctl_elem_id_get_name(id))) {
      outvolId=snd_ctl_elem_id_get_numid(id);
      outvolCount=count;
printf("Playback Volume id=%d [%d]\n", outvolId, count);
    } else if (!strcmp("Line Playback Volume", snd_ctl_elem_id_get_name(id))) {
      masterId=snd_ctl_elem_id_get_numid(id);
printf("Line Volume id=%d\n", masterId);
    } else if (!strcmp("Line Capture Volume", snd_ctl_elem_id_get_name(id))) {
      involId=snd_ctl_elem_id_get_numid(id);
printf("Capture Volume id=%d [%d]\n", involId, count);
    } else if (!strcmp("Line Playback Switch (-10dBV)", snd_ctl_elem_id_get_name(id))) {
      p4OutId=numid;
printf("Playback nominal id=%d [%d]\n", p4OutId, count);
    } else if (!strcmp("Line Capture Switch (-10dBV)", snd_ctl_elem_id_get_name(id))) {
      p4InId=numid;
printf("Capture nominal id=%d [%d]\n", p4InId, count);
    } else if (!strcmp("Digital mode Switch", snd_ctl_elem_id_get_name(id))) {
      dmodeId=numid;
      items=snd_ctl_elem_info_get_items(info);
      ndmodes=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info)) < 0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(dmodeName[item], snd_ctl_elem_info_get_item_name(info), 63);
        dmodeName[item][63]=0;
printf("Digital Mode id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("Clock source Switch", snd_ctl_elem_id_get_name(id))) {
      clocksrcId=numid;
      items=snd_ctl_elem_info_get_items(info);
      nclocksrc=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info))<0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(clocksrcName[item], snd_ctl_elem_info_get_item_name(info), 63);
        clocksrcName[item][63]=0;
printf("Clock source id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("S/PDIF mode Switch", snd_ctl_elem_id_get_name(id))) {
      spdifmodeId=numid;
      items=snd_ctl_elem_info_get_items(info);
      nspdifmodes=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info)) < 0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(spdifmodeName[item], snd_ctl_elem_info_get_item_name(info), 63);
        spdifmodeName[item][63]=0;
printf("S/PDIF Mode id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("Dithering Switch", snd_ctl_elem_id_get_name(id))) {
      ditherId=numid;
printf("Dithering switch id=%d\n", numid);
    } else if (!strcmp("VU-meters Switch", snd_ctl_elem_id_get_name(id))) {
      vuswitchId=numid;
printf("VU-meter switch id=%d\n", numid);
    } else if (!strcmp("VU-meters", snd_ctl_elem_id_get_name(id))) {
      vumetersId=numid;
      metersStream=snd_ctl_elem_info_get_dimension(info, 0);
      metersNumber=snd_ctl_elem_info_get_dimension(info, 1);
      metersTypes=snd_ctl_elem_info_get_dimension(info, 2);
printf("VU-meters id=%d\n", numid);
    } else if (!strcmp("Channels info", snd_ctl_elem_id_get_name(id))) {
      channelsId=numid;
printf("Channels info id=%d\n", numid);
    }
  }

  GetChannels();
printf("nIn=%d fdIn=%d nLOut=%d nPOut=%d fdOut=%d\n", nIn,fdIn,nLOut,nPOut, fdOut);

  if (FirstMixerId>=0 && (mixerInputs!=nIn || mixerOutputs!=nLOut)) {
    printf("** Warning ** Mixer/channels mismatch !!  nIn=%d nIn=%d    nLOut=%d nLOut=%d\n", nIn, mixerInputs, nLOut, mixerOutputs);
    return(1);
  }
  if (outvolCount!=nLOut) {
    printf("** Warning ** mismatch !!  outvolCount=%d nLOut=%d\n", outvolCount, nLOut);
    return(1);
  }
  if (FirstVmixerId>=0) {
    if (vmixerPipes!=nPOut || vmixerOutputs!=fdOut) {
      printf("error: vmixer/channels mismatch  vmp=%d nvo=%d    vmo=%d fdo=%d !!\n", vmixerPipes, nPOut, vmixerOutputs, fdOut);
      return(1);
    }
    if (vmixerOutputs>2) {
      printf("error: more than 2 vmixer outputs are not supported.\n");
      return(1);
    }
  }
  
  //@ Assumes all mixer and vmixer controls are contiguous
  for (o=0; o<nLOut; o++) {
    for (i=0; i<nIn; i++) {
      Mixer[o][i].id=FirstMixerId++;
    }
  }

  for (i=0; i<nLOut; i++) {
    for (o=0; o<nPOut; o++) {
      Vmixer[i][o].id=FirstVmixerId++;
    }
  }

  snd_hctl_close(handle);
  return(0);
}



int main(int argc, char *argv[]) {
  gchar str[256];
  GtkWidget *hbox, *vbox;
  GtkWidget *hbmain;
  GtkWidget *vbsel, *frame, *button;
  GtkWidget *label, *menu, *menuitem;
  GSList *in_wgroup, *out_wgroup;
  int err, i, o, n, cardnum;
  char hwname[8], cardname[32], cardId[16], load, save;
  snd_ctl_card_info_t *hw_info;

  load=save=1;

  // Scans all installed cards
  snd_ctl_card_info_alloca(&hw_info);
  cardnum=-1;
  ctlhandle=0;

  while (snd_card_next(&cardnum)>=0 && cardnum>=0) {
    sprintf(hwname, "hw:%d", cardnum);
    if ((err=snd_ctl_open(&ctlhandle, hwname, 0))<0) {
      printf("snd_ctl_open(%s) Error: %s\n", hwname, snd_strerror(err));
      continue;
    }
    if ((err=snd_ctl_card_info(ctlhandle, hw_info))>=0) {
      if (!strcmp(snd_ctl_card_info_get_driver(hw_info), "Echoaudio")) {
        strncpy(card, hwname, 7);
        hwname[7]=0;
        strncpy(cardname, snd_ctl_card_info_get_name(hw_info), 31);
        cardname[31]=0;
        strncpy(cardId, snd_ctl_card_info_get_name(hw_info), 15);
        cardId[15]=0;
        printf("Card found: %s  (%s)\n", snd_ctl_card_info_get_longname(hw_info), hwname);
/*printf("snd_ctl_card_info_get_card = %d\n", snd_ctl_card_info_get_card(hw_info));
printf("snd_ctl_card_info_get_id = %s\n", snd_ctl_card_info_get_id(hw_info));
printf("snd_ctl_card_info_get_driver = %s\n", snd_ctl_card_info_get_driver(hw_info));
printf("snd_ctl_card_info_get_name = %s\n", snd_ctl_card_info_get_name(hw_info));
printf("snd_ctl_card_info_get_longname = %s\n", snd_ctl_card_info_get_longname(hw_info));
printf("snd_ctl_card_info_get_mixername = %s\n", snd_ctl_card_info_get_mixername(hw_info));
printf("snd_ctl_card_info_get_components = %s\n", snd_ctl_card_info_get_components(hw_info));*/
        break;
      }
    } else {
      printf("snd_ctl_card_info(%s) Error: %s\n", hwname, snd_strerror(err));
    }
    snd_ctl_close(ctlhandle);
    ctlhandle=0;
  }

  if (!ctlhandle) {
    printf("No Echoaudio cards found, sorry.\n");
    return(0);
  }

  // Reads available controls
  if (OpenControls(card, cardname))
    exit(1);

  mouseButton=0;
  Gang=1;
  memset(LineOutVol, 0, sizeof(LineOutVol));
  memset(PcmOutVol, 0, sizeof(LineOutVol));
  memset(LineInVol, 0, sizeof(LineOutVol));
  Mainw_geom.st=NOPOS;
  PVw_geom.st=NOPOS;
  LVw_geom.st=NOPOS;
  VUw_geom.st=NOPOS;
  Mixerw_geom.st=NOPOS;
  VUwindow=GMwindow=0;
  GMixerSection.Monitor=nIn-1;  // The correct value is set by Digital_mode_activate()
  GMixerSection.Vmixer=nIn-1;//@@
  GMixerSection.Line=GMixerSection.Vmixer+1;
  InitMixer();

  //@@ check the values
  if (load) {
    FILE *f;
    snprintf(str, 255, "%s/.Emixer_%s", getenv("HOME"), cardId);
    str[255]=0;
    if (f=fopen(str, "r")) {
      str[255]=0;
      while (fgets(str, 255, f)) {
        if (!strncmp("LineOut ", str, 8)) {
          sscanf(str+8, "%d %d", &o, &n);
          if (o>=0 && o<nLOut)
            LineOutVol[o]=n;
        } else if (!strncmp("LineIn ", str, 7)) {
          sscanf(str+7, "%d %d", &i, &n);
          if (i>=0 && i<nIn)
            LineInVol[i]=n;
        } else if (!strncmp("PcmOut ", str, 7)) {
          sscanf(str+7, "%d %d", &o, &n);
          if (o>=0 && o<nPOut)
            PcmOutVol[o]=n;
        } else if (!strncmp("Mixer ", str, 6)) {
          sscanf(str+6, "%d %d %d", &o, &i, &n);
          if (o>=0 && o<nLOut && i>=0 && i<nIn)
            Mixer[o][i].val=n;
        } else if (!strncmp("MainWindow ", str, 11)) {
          sscanf(str+11, "%d %d %d %d", &Mainw_geom.x, &Mainw_geom.y, &Mainw_geom.w, &Mainw_geom.h);
        } else if (!strncmp("VUmetersWindow ", str, 15)) {
          sscanf(str+15, "%d %d %d", &VUw_geom.x, &VUw_geom.y, &VUw_geom.st);
        } else if (!strncmp("GfxMixerWindow ", str, 15)) {
          sscanf(str+15, "%d %d %d", &GMw_geom.x, &GMw_geom.y, &GMw_geom.st);
        } else if (!strncmp("PcmVolumeWindow ", str, 16)) {
          sscanf(str+16, "%d %d %d %d %d", &PVw_geom.x, &PVw_geom.y, &PVw_geom.w, &PVw_geom.h, &PVw_geom.st);
        } else if (!strncmp("LineVolumeWindow ", str, 17)) {
          sscanf(str+17, "%d %d %d %d %d", &LVw_geom.x, &LVw_geom.y, &LVw_geom.w, &LVw_geom.h, &LVw_geom.st);
        } else if (!strncmp("MixerWindow ", str, 12)) {
          sscanf(str+12, "%d %d %d %d %d", &Mixerw_geom.x, &Mixerw_geom.y, &Mixerw_geom.w, &Mixerw_geom.h, &Mixerw_geom.st);
        }
      }
    }
  }

  gtk_init(&argc, &argv);
  fnt=gdk_font_load("-misc-fixed-medium-r-*-*-10-*-*-*-*-*-*-*");
  if (!fnt) {
    printf("Cannot find the font\n");
    exit(1);
  }



/* ********** Main window ********** */

  Mainwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (Mainwindow), VERSION);
  gtk_signal_connect(GTK_OBJECT(Mainwindow), "delete_event", GTK_SIGNAL_FUNC(Mainwindow_delete), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(Mainwindow), BORDER);
  gtk_widget_show(Mainwindow);
  if (Mainw_geom.x!=NOPOS) {
    gtk_widget_set_uposition(Mainwindow, Mainw_geom.x, Mainw_geom.y);
    gtk_window_set_default_size(GTK_WINDOW(Mainwindow), Mainw_geom.w, Mainw_geom.h);
  }

  hbmain=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain);
  gtk_container_add(GTK_CONTAINER(Mainwindow), hbmain);

  if (p4InId) {
    // Consumer/professional analog input switches
    frame=gtk_frame_new("Input +4dBu");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdIn; i++) {
      sprintf(str, "%d", i);
      p4dbuIn[i]=gtk_toggle_button_new_with_label(str);
      gtk_box_pack_start(GTK_BOX(hbox), p4dbuIn[i], TRUE, FALSE, 1);
      gtk_widget_show(p4dbuIn[i]);
      gtk_signal_connect(GTK_OBJECT(p4dbuIn[i]), "toggled", GTK_SIGNAL_FUNC(Nominal_level_toggled), (gpointer)i);
    }
    InitNominalLevelGUI(p4InId);
  }


  if (p4OutId) {
    // Consumer/professional analog output switches
    frame=gtk_frame_new("Output +4dBu");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdOut; i++) {
      sprintf(str, "%d", i);
      p4dbuOut[i]=gtk_toggle_button_new_with_label(str);
      gtk_box_pack_start(GTK_BOX(hbox), p4dbuOut[i], TRUE, FALSE, 1);
      gtk_widget_show(p4dbuOut[i]);
      gtk_signal_connect(GTK_OBJECT(p4dbuOut[i]), "toggled", GTK_SIGNAL_FUNC(Nominal_level_toggled), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
    }
    InitNominalLevelGUI(p4OutId);
  }


  if (dmodeId && ndmodes>1) {
    // Digital mode switch
    frame=gtk_frame_new("Digital mode");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    dmodeOpt=gtk_option_menu_new();
    gtk_widget_show(dmodeOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<ndmodes; i++) {
      menuitem=gtk_menu_item_new_with_label(dmodeName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", Digital_mode_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(dmodeOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), dmodeOpt, FALSE, FALSE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(dmodeOpt), dmodeVal=GetEnum(dmodeId));
  }


  if (clocksrcId && nclocksrc>1) {
    // Clock source switch
    frame=gtk_frame_new("Clock source");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    clocksrcOpt=gtk_option_menu_new();
    gtk_widget_show(clocksrcOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<nclocksrc; i++) {
      menuitem=gtk_menu_item_new_with_label(clocksrcName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", Clock_source_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(clocksrcOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), clocksrcOpt, FALSE, FALSE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal=GetEnum(clocksrcId));
  }


  if (spdifmodeId && nspdifmodes>1) {
    // S/PDIF mode switch
    frame=gtk_frame_new("S/PDIF mode");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    spdifmodeOpt=gtk_option_menu_new();
    gtk_widget_show(spdifmodeOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<nspdifmodes; i++) {
      menuitem=gtk_menu_item_new_with_label(spdifmodeName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", SPDIF_mode_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(spdifmodeOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), spdifmodeOpt, TRUE, TRUE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(spdifmodeOpt), spdifmodeVal=GetEnum(spdifmodeId));
  }


  if (ditherId) {
    // 16 bits digital input dithering switch
    frame=gtk_frame_new("Dither");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    ditherToggle=gtk_toggle_button_new_with_label("On");
    gtk_widget_show(ditherToggle);
    gtk_box_pack_start(GTK_BOX(hbox), ditherToggle, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(ditherToggle), "toggled", Dithering_toggled, (gpointer)0);
    InitDitheringGUI(ditherId);
  }


  if (involId || outvolId) {
    // Gang button
    frame=gtk_frame_new("Gang");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button=gtk_toggle_button_new_with_label("On");
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 1);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "toggled", Gang_button_toggled, 0);
  }


  frame=gtk_frame_new("Windows");
  gtk_widget_show(frame);
  gtk_box_pack_end(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
  hbox=gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox);
  gtk_container_add(GTK_CONTAINER(frame), hbox);

  if (vumetersId && vuswitchId) {
    button=gtk_button_new_with_label("VU");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", VUmeters_button_click, 0);
    if (VUw_geom.st==1)
      VUmeters_button_click(button, 0);
  }
  if (Mixer[0][0].id>=0) {
    button=gtk_button_new_with_label("GrMix");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GMixer_button_click, 0);
    if (GMw_geom.st==1)
      GMixer_button_click(button, 0);

    button=gtk_button_new_with_label("Mixer");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", Mixer_button_click, 0);
  }
  if (outvolId) {
    button=gtk_button_new_with_label("PCM");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", PV_button_click, 0);
  }

  button=gtk_button_new_with_label("Line");
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 1);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", LV_button_click, 0);



/* ********** PCM volume window ********** */

  PVwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (PVwindow), "PCM");
  gtk_signal_connect(GTK_OBJECT(PVwindow), "delete_event", GTK_SIGNAL_FUNC(window_close), (gpointer)&PVw_geom);
  gtk_container_set_border_width(GTK_CONTAINER(PVwindow), BORDER);
  if (PVw_geom.st!=NOPOS) {
    if (PVw_geom.st==1)
      gtk_widget_show(PVwindow);
    gtk_widget_set_uposition(PVwindow, PVw_geom.x, PVw_geom.y);
    gtk_window_set_default_size(GTK_WINDOW(PVwindow), PVw_geom.w, PVw_geom.h);
  }

  hbmain=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain);
  gtk_container_add(GTK_CONTAINER(PVwindow), hbmain);


  if (outvolId) {
    // PCM Output volume widgets
    frame=gtk_frame_new("PCM Output volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nPOut; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdOut)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdOut);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      outvoladj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      outvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(outvoladj[i]));
      gtk_widget_show(outvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), outvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(outvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(outvoladj[i]), "value_changed", GTK_SIGNAL_FUNC(PCM_volume_changed), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
      // Value label
      outvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(outvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), outvol_lbl[i], FALSE, FALSE, 0);
    }
    gtk_widget_set_usize(GTK_WIDGET(outvol[0]), 0, 170);        // Set minimum y size
    InitPCMVolumeGUI(OUTPUT, load);
  }


/* ********** Line volume window ********** */

  LVwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (LVwindow), "Line");
  gtk_signal_connect(GTK_OBJECT(LVwindow), "delete_event", GTK_SIGNAL_FUNC(window_close), (gpointer)&LVw_geom);
  gtk_container_set_border_width(GTK_CONTAINER(LVwindow), BORDER);
  if (LVw_geom.st!=NOPOS) {
    if (LVw_geom.st==1)
      gtk_widget_show(LVwindow);
    gtk_widget_set_uposition(LVwindow, LVw_geom.x, LVw_geom.y);
    gtk_window_set_default_size(GTK_WINDOW(LVwindow), LVw_geom.w, LVw_geom.h);
  }

  hbmain=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain);
  gtk_container_add(GTK_CONTAINER(LVwindow), hbmain);

  // Input volume widgets
  if (involId) {
    frame=gtk_frame_new("Analog input volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdIn; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      sprintf(str, "%d", i);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      involadj[i]=gtk_adjustment_new(0, ECHOGAIN_MININP, ECHOGAIN_MAXINP, LEVELSTEP/2, LEVELSTEP*3, 0);
      invol[i]=gtk_vscale_new(GTK_ADJUSTMENT(involadj[i]));
      gtk_widget_show(invol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), invol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(invol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(involadj[i]), "value_changed", GTK_SIGNAL_FUNC(PCM_volume_changed), (gpointer)i);
      // Value label
      invol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(invol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), invol_lbl[i], FALSE, FALSE, 0);
    }
    gtk_widget_set_usize(GTK_WIDGET(invol[0]), 0, 170);         // Set minimum y size
    InitPCMVolumeGUI(INPUT, load);
  }


  // Line output volume widgets
  if (1) {
    frame=gtk_frame_new("Line output volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nLOut; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdOut)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdOut);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      mstadj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      mstvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(mstadj[i]));
      gtk_widget_show(mstvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mstvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(mstvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(mstadj[i]), "value_changed", GTK_SIGNAL_FUNC(Line_volume_changed), (gpointer)i);
      // Value label
      mstvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(mstvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mstvol_lbl[i], FALSE, FALSE, 0);
    }
    gtk_widget_set_usize(GTK_WIDGET(mstvol[0]), 0, 170);        // Set minimum y size
    InitLineVolumeGUI();
  }



/* ********** Mixer window ********** */

  Mixerwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (Mixerwindow), "Monitor mixer");
  gtk_signal_connect(GTK_OBJECT(Mixerwindow), "delete_event", GTK_SIGNAL_FUNC(window_close), (gpointer)&Mixerw_geom);
  gtk_container_set_border_width(GTK_CONTAINER(Mixerwindow), BORDER);
  if (Mixerw_geom.st!=NOPOS) {
    if (Mixerw_geom.st==1)
      gtk_widget_show(Mixerwindow);
    gtk_widget_set_uposition(Mixerwindow, Mixerw_geom.x, Mixerw_geom.y);
    gtk_window_set_default_size(GTK_WINDOW(Mixerwindow), Mixerw_geom.w, Mixerw_geom.h);
//    gdk_window_move_resize(Mixerwindow->window, Mixerw_geom.x, Mixerw_geom.y, Mixerw_geom.w, Mixerw_geom.h);
/*    gtk_widget_set_usize(Mixerwindow, Mixerw_geom.w, Mixerw_geom.h);
    gtk_widget_set_usize(Mixerwindow, -1, -1);*/
  }

  hbmain=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain);
  gtk_container_add(GTK_CONTAINER(Mixerwindow), hbmain);


  // Mixer
  if (Mixer[0][0].id>=0) {
#ifdef REVERSE
    // Mixer volume widgets
    frame=gtk_frame_new("Mixer input levels");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nIn; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdIn)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdIn);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      mixeradj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      mixvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(mixeradj[i]));
      gtk_widget_show(mixvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(mixvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(mixvol[i]), "grab_focus", GTK_SIGNAL_FUNC(Monitor_involume_clicked), (gpointer)i);
      gtk_signal_connect(GTK_OBJECT(mixeradj[i]), "value_changed", GTK_SIGNAL_FUNC(Monitor_volume_changed), (gpointer)i);
      // Value label
      mixvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(mixvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol_lbl[i], FALSE, FALSE, 0);
    }
    gtk_widget_set_usize(GTK_WIDGET(mixvol[0]), 0, 170);        // Set minimum y size

    // Output channel selectors
    frame=gtk_frame_new("Mixer output");
    gtk_widget_show(frame);
    vbsel=gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    gtk_widget_show(vbsel);
    gtk_container_add(GTK_CONTAINER(frame), vbsel);

    out_wgroup=0;
    for (i=n=0; i<nLOut; i++) {
      if (i<fdOut)
        sprintf(str, "An-%d", i);
      else
        sprintf(str, "Di-%d", i-fdOut);
      if (i)
        out_wgroup=gtk_radio_button_group(GTK_RADIO_BUTTON(mixout[i-1]));
      mixout[i]=gtk_radio_button_new_with_label(out_wgroup, str);
      gtk_widget_show(mixout[i]);
      gtk_box_pack_start(GTK_BOX(vbsel), mixout[i], FALSE, FALSE, 0);
      gtk_signal_connect(GTK_OBJECT(mixout[i]), "toggled", GTK_SIGNAL_FUNC(Output_selector_clicked), (gpointer)i);
    }
    InputChannel=0;
    OutputChannel=-1;
    Output_selector_clicked(0, 0);

#else // REVERSE

    // Input channel selectors
    frame=gtk_frame_new("Mixer input");
    gtk_widget_show(frame);
    vbsel=gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, FALSE, FALSE, 0);
    gtk_widget_show(vbsel);
    gtk_container_add(GTK_CONTAINER(frame), vbsel);

    in_wgroup=0;
    for (i=n=0; i<nIn; i++) {
      if (i<fdIn)
        sprintf(str, "An-%d", i);
      else
        sprintf(str, "Di-%d", i-fdIn);
      if (i)
        in_wgroup=gtk_radio_button_group(GTK_RADIO_BUTTON(mixin[i-1]));
      mixin[i]=gtk_radio_button_new_with_label(in_wgroup, str);
      gtk_widget_show(mixin[i]);
      gtk_box_pack_start(GTK_BOX(vbsel), mixin[i], FALSE, FALSE, 0);
      gtk_signal_connect(GTK_OBJECT(mixin[i]), "toggled", GTK_SIGNAL_FUNC(Input_selector_clicked), (gpointer)i);
    }

    // Mixer volume widgets
    frame=gtk_frame_new("Mixer output levels");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nLOut; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdOut)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdOut);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      mixeradj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      mixvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(mixeradj[i]));
      gtk_widget_show(mixvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(mixvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(mixvol[i]), "grab_focus", GTK_SIGNAL_FUNC(Monitor_outvolume_clicked), (gpointer)i);
      gtk_signal_connect(GTK_OBJECT(mixeradj[i]), "value_changed", GTK_SIGNAL_FUNC(Monitor_volume_changed), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
      // Value label
      mixvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(mixvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol_lbl[i], FALSE, FALSE, 0);
    }
    gtk_widget_set_usize(GTK_WIDGET(mixvol[0]), 0, 170);        // Set minimum y size
    InputChannel=-1;
    OutputChannel=0;
    Input_selector_clicked(0, 0);
#endif
  }



/* ********** GTK-main ********** */

  Digital_mode_activate(dmodeOpt, (gpointer)dmodeVal);
  //SetSensitivity(ADATmode());
  gtk_widget_show(Mainwindow);
  gtk_main();

  if (save) {
    FILE *f;
    if (snprintf(str, 255, "%s/.Emixer_%s", getenv("HOME"), cardId)>0) {
      str[255]=0;
      if (f=fopen(str, "w")) {
        fprintf(f, "-- LineOut <channel> <gain>\n");
        for (i=0; i<nLOut; i++)
          fprintf(f, "LineOut %2d %d\n", i, LineOutVol[i]);
        fprintf(f, "-- LineIn <channel> <gain>\n");
        for (i=0; i<nIn; i++)
          fprintf(f, "LineIn %2d %d\n", i, LineInVol[i]);
        fprintf(f, "-- PcmOut <channel> <gain>\n");
        for (i=0; i<nPOut; i++)
          fprintf(f, "PcmOut %2d %d\n", i, PcmOutVol[i]);
        fprintf(f, "-- Mixer <output> <input> <gain>\n");
        for (o=0; o<nLOut; o++)
          for (i=0; i<nIn; i++)
            fprintf(f, "Mixer %2d %2d %d\n", o, i, Mixer[o][i].val);
        fprintf(f, "-- xxWindow <x> <y> <width> <height> <visible>\n");
        fprintf(f, "MainWindow %d %d %d %d\n", Mainw_geom.x, Mainw_geom.y, Mainw_geom.w, Mainw_geom.h);
        if (VUwindow)
          gdk_window_get_root_origin(VUwindow->window, &VUw_geom.x, &VUw_geom.y);
        fprintf(f, "VUmetersWindow %d %d %d\n", VUw_geom.x, VUw_geom.y, VUw_geom.st);
        if (GMwindow)
          gdk_window_get_root_origin(GMwindow->window, &VUw_geom.x, &VUw_geom.y);
        fprintf(f, "GfxMixerWindow %d %d %d\n", GMw_geom.x, GMw_geom.y, GMw_geom.st);
        if (PVwindow->window) {
          gdk_window_get_root_origin(PVwindow->window, &PVw_geom.x, &PVw_geom.y);
          gdk_window_get_size(PVwindow->window, &PVw_geom.w, &PVw_geom.h);
        }
        fprintf(f, "PcmVolumeWindow %d %d %d %d %d\n", PVw_geom.x, PVw_geom.y, PVw_geom.w, PVw_geom.h, !!GTK_WIDGET_VISIBLE(PVwindow));
        if (LVwindow->window) {
          gdk_window_get_root_origin(LVwindow->window, &LVw_geom.x, &LVw_geom.y);
          gdk_window_get_size(LVwindow->window, &LVw_geom.w, &LVw_geom.h);
        }
        fprintf(f, "LineVolumeWindow %d %d %d %d %d\n", LVw_geom.x, LVw_geom.y, LVw_geom.w, LVw_geom.h, !!GTK_WIDGET_VISIBLE(LVwindow));
        if (Mixerwindow->window) {
          gdk_window_get_root_origin(Mixerwindow->window, &Mixerw_geom.x, &Mixerw_geom.y);
          gdk_window_get_size(Mixerwindow->window, &Mixerw_geom.w, &Mixerw_geom.h);
        }
        fprintf(f, "MixerWindow %d %d %d %d %d\n", Mixerw_geom.x, Mixerw_geom.y, Mixerw_geom.w, Mixerw_geom.h, !!GTK_WIDGET_VISIBLE(Mixerwindow));
        fprintf(f, "\n");
        fclose(f);
      }
    }
  }

  if (VUwindow) {
    SetVUmeters(0);
    gtk_timeout_remove(VUtimer);
  }
  if (GMwindow) {
    SetVUmeters(0);
    gtk_timeout_remove(Mixtimer);
  }
  snd_ctl_close(ctlhandle);
  return(0);
}

