/*
 *  ALSA mixer console for Echoaudio soundcards.
 *  Copyright (C) 2003 Giuliano Pochini <pochini@shiny.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define VERSION "Echoaudio console v0.8.2"

/*
 Uncomment this to change the VUmeters window in a complete and interactive
 mixer. On the left you can see the levels of the input channels. Each cell
 has a volume know and a VUmeter that displays the level of the respective
 input channel modified by its volume setting. The bottom line shows the
 levels of the output channels.
*/
//#define MATRIX


/*
 This option changes the mixer so that you can select the output channel on
 the mixer in place of the input channel.
*/
//#define REVERSE



#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <alsa/asoundlib.h>

#define DSP_TO_GENERIC(Value) (Value << 8)
#undef TRUE
#undef FALSE
#include "EchoGalsXface.h"
#include "MixerXface.h"

#define LEVELSTEP DSP_TO_GENERIC(1)      // 1dB
#define MAXMODES 16 //@fare dinamico
#define INPUT 0
#define OUTPUT 1

#define BARWIDTH 5   // Width of VU-meters bars
#define BARSEP 2     // Spacing between VU-meters bars
#define BORDER 6     // Inner window border
#define SPACING 8    // Spacing of control sections

char card[64];
char dmodeName[MAXMODES][64], clocksrcName[MAXMODES][64], spdifmodeName[MAXMODES][64];
int nOut, nIn, fdIn, fdOut, nVOut;
int ndmodes, nclocksrc, nspdifmodes, mixerInputs, mixerOutputs, vmixerPipes, vmixerOutputs;
int InputChannel, OutputChannel, Gang;
int involId, outvolId, masterId, vmixerId, p4InId, p4OutId, dmodeId, clocksrcId, spdifmodeId, vuswitchId, vumetersId, channelsId, ditherId;
int outvolCount;
int mouseY, mouseButton;
int dmodeVal, clocksrcVal, spdifmodeVal;
int VUwidth, VUheight, Mixwidth, Mixheight;
struct mixel {
  int id;
  int val;
} Mixer[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOINPUTS], Vmixer[ECHO_MAXAUDIOOUTPUTS][ECHO_MAXAUDIOOUTPUTS];
snd_ctl_t *ctlhandle;

#define ECHO_MAXAUDIO_IOS ECHO_MAXAUDIOINPUTS  // It's the maximum between inputs an outputs

GtkObject *mixeradj[ECHO_MAXAUDIOOUTPUTS], *involadj[ECHO_MAXAUDIOOUTPUTS], *outvoladj[ECHO_MAXAUDIOOUTPUTS], *vmixeradj[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *invol[ECHO_MAXAUDIOOUTPUTS], *invol_lbl[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *outvol[ECHO_MAXAUDIOOUTPUTS], *outvol_lbl[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *mixvol[ECHO_MAXAUDIOOUTPUTS], *mixvol_lbl[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *vmixervol[ECHO_MAXAUDIOOUTPUTS], *vmixer_lbl[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *p4dbuOut[ECHO_MAXAUDIOOUTPUTS], *p4dbuIn[ECHO_MAXAUDIOINPUTS];
GtkWidget *mixin[ECHO_MAXAUDIOINPUTS], *mixout[ECHO_MAXAUDIOOUTPUTS];  // In and out selectors
GtkWidget *dmodeOpt, *clocksrcOpt, *spdifmodeOpt, *ditherToggle;
GtkWidget *window, *VUwindow, *Mixwindow;
GtkWidget *VUs[ECHO_MAXAUDIOOUTPUTS];
GtkWidget *VUdarea, *Mixdarea;
gint VUtimer, Mixtimer;

GdkGC *gc=0;
static GdkPixmap *VUpixmap = NULL;
static GdkPixmap *Mixpixmap = NULL;
GdkFont *fnt;




int CountBits(int n) {
  int c;

  c=0;
  while (n) {
    c++;
    n&=(n-1);
  }
  return(c);
}



int ADATmode() {
  return(!strncmp(dmodeName[dmodeVal], "ADAT", 4));
}



// Write an enumerated ALSA control
int SetEnum(int numid, int val) {
  int err;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, numid);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_enumerated(control, 0, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
  return(err);
}



// Read an enumerated ALSA control
int GetEnum(int numid) {
  int err, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, numid);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control)) < 0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  val=snd_ctl_elem_value_get_enumerated(control, 0);
  return(val);
}



// Turn VU-meters on/off
void SetVUmeters(int onoff) {
  static signed char oncount=0;
  int err;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (onoff)
    oncount++;
  else
    if (--oncount<0)
      oncount=0;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, vuswitchId);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, !!oncount);
  if ((err=snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}



void GetVUmeters(int *InLevel, int *InPeak, int *OutLevel, int *OutPeak) {
  int err, i;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_id_set_numid(id, vumetersId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err = snd_ctl_elem_read(ctlhandle, control)) < 0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
    return;
  }
  for (i=0; i<nIn; i++) {
    InLevel[i]=snd_ctl_elem_value_get_integer(control, i);
    InPeak[i]=snd_ctl_elem_value_get_integer(control, ECHO_MAXAUDIOINPUTS+i);
  }
  for (i=0; i<nOut; i++) {
    OutLevel[i]=snd_ctl_elem_value_get_integer(control, ECHO_MAXAUDIOINPUTS*2+i);
    OutPeak[i]=snd_ctl_elem_value_get_integer(control, ECHO_MAXAUDIOINPUTS*2+ECHO_MAXAUDIOOUTPUTS+i);
  }
}



#ifdef REVERSE

// Enable/disable widgets that control ADAT digital channels
void SetSensitivity(int enable) {
  int i;

  for (i=fdOut+2; i<nOut; i++) {
    if (!enable && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mixout[i])))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixout[0]), TRUE);
    gtk_widget_set_sensitive(mixout[i], enable);
    gtk_widget_set_sensitive(outvol_lbl[i], enable);
    gtk_widget_set_sensitive(outvol[i], enable);
  }
  for (i=fdIn+2; i<nIn; i++) {
    gtk_widget_set_sensitive(mixvol_lbl[i], enable);
    gtk_widget_set_sensitive(mixvol[i], enable);
  }
  if (!enable && InputChannel>=fdIn+2)
    InputChannel=0;
}

#else

// Enable/disable widgets that control ADAT digital channels
void SetSensitivity(int enable) {
  int i;

  for (i=fdIn+2; i<nIn; i++) {
    if (!enable && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mixin[i])))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixin[0]), TRUE);
    gtk_widget_set_sensitive(mixin[i], enable);
  }
  for (i=fdOut+2; i<nOut; i++) {
    gtk_widget_set_sensitive(mixvol_lbl[i], enable);
    gtk_widget_set_sensitive(mixvol[i], enable);
    gtk_widget_set_sensitive(outvol_lbl[i], enable);
    gtk_widget_set_sensitive(outvol[i], enable);
  }
  if (!enable && OutputChannel>=fdOut+2)
    OutputChannel=0;
}

#endif


// At startup this functions reads the current volume levels and sets the widgets accordingly.
void InitVolumeGUI(int inout) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, ch, channels, volId, val, fix;
  char str[16], *fmt;
  GtkObject **adj;
  GtkWidget **label;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);

  if (inout==INPUT) {
    channels=fdIn;
    label=invol_lbl;
    fmt="%+4.1f";
    adj=involadj;
    volId=involId;
    fix=ECHOGAIN_MININP+ECHOGAIN_MAXINP;
  } else {
    channels=outvolCount;
    label=outvol_lbl;
    fmt="%+3.0f";
    adj=outvoladj;
    volId=outvolId;
    fix=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT;
  }

  snd_ctl_elem_id_set_numid(id, volId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
    return;
  }
  for (ch=0; ch<channels; ch++) {
    val=snd_ctl_elem_value_get_integer(control, ch);
    if (val==ECHOGAIN_MINOUT)
      strcpy(str, "mute");
    else
      sprintf(str, fmt, (float)val/256.0);
    gtk_label_set_text(GTK_LABEL(label[ch]), str);
    val=fix-val;     // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(adj[ch]), (gfloat)val);
  }
}



// At startup this functions reads the current nominal levels and sets the switches accordingly.
void InitNominalLevelGUI(int numid) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, i, n;
  GtkWidget **w;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  if (numid==p4InId) {
    snd_ctl_elem_id_set_numid(id, p4InId);
    n=fdIn;
    w=p4dbuIn;
  } else {
    snd_ctl_elem_id_set_numid(id, p4OutId);
    n=fdOut;
    w=p4dbuOut;
  }
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  for (i=0; i<n; i++)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w[i]), !snd_ctl_elem_value_get_integer(control, i)); // FALSE is +4
}



// At startup this functions reads if the dithering is enabled sets the button accordingly.
void InitDitheringGUI(int numid) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, ditherId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ditherToggle), snd_ctl_elem_value_get_integer(control, 0));
}



// Read all mixer settings.
void InitMixer(void) {
  int err, in, out;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);

  for (out=0; out<nOut; out++) {
    for (in=0; in<nIn; in++) {
      snd_ctl_elem_id_set_numid(id, Mixer[out][in].id);
      snd_ctl_elem_value_set_id(control, id);
      if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
        printf("Control %s element read error: %s\n", card, snd_strerror(err));
      }
      Mixer[out][in].val=snd_ctl_elem_value_get_integer(control, 0);
    }
  }
}



#define XCELLBORDER 2
#define YCELLBORDER 2
#define XCELLDIM 20
#define YCELLDIM 32
#define XCELLTOT (1+XCELLBORDER*2+XCELLDIM)
#define YCELLTOT (1+YCELLBORDER*2+YCELLDIM)
#define XVOLUME 6
#define XMETER 14

// Draw the matrix mixer
gint UpdateMixer(gpointer data) {
  GdkRectangle update_rect;
  int InLevel[ECHO_MAXAUDIOINPUTS];
  int InPeak[ECHO_MAXAUDIOINPUTS];
  int OutLevel[ECHO_MAXAUDIOOUTPUTS];
  int OutPeak[ECHO_MAXAUDIOOUTPUTS];
  static int InClip[ECHO_MAXAUDIOINPUTS];
  static int OutClip[ECHO_MAXAUDIOOUTPUTS];
  char str[8];
  int i, o, x, dB, db, inchannels, outchannels;
  GdkColor Grid={0x9694C4, 0, 0, 0};
  GdkColor Bars={0x00FF00, 0, 0, 0};
  GdkColor Bars1={0x000000, 0, 0, 0};
  GdkColor Peak={0x1BABFF, 0, 0, 0};
  GdkColor Level={0xC0B000, 0, 0, 0};
  GdkColor Hilight={0x000078, 0, 0, 0};

#define MINXI 30
#define MINXO 132
#define MINY 20

  if (ADATmode()) {
    inchannels=nIn;
    outchannels=nOut;
  } else {
    inchannels=fdIn+2;
    outchannels=fdOut+2;
  }

  if (!Mixpixmap)
    return(TRUE);

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = Mixwidth;
  update_rect.height = Mixheight;
  GetVUmeters(InLevel, InPeak, OutLevel, OutPeak);

  if (!gc) {
    gc=gdk_gc_new(gtk_widget_get_parent_window(Mixdarea));
    for (i=0; i<nIn; i++)
      InClip[i]=0;
    for (i=0; i<nOut; i++)
      OutClip[i]=0;
  }

  gdk_draw_rectangle(Mixpixmap, Mixdarea->style->black_gc, TRUE, 0, 0, Mixwidth, Mixheight);

/*
  gdk_draw_rectangle(Mixpixmap, Mixdarea->style->black_gc, TRUE, 0, 0, Mixwidth, Mixheight);

  gdk_gc_set_foreground(gc, &Grid);
  gdk_gc_set_font(gc, fnt);
  gdk_draw_string(Mixpixmap, fnt, gc, 2, MINY-12+4, "  dB");
  for (i=0; i<=120; i+=12) {
    sprintf(str, "%4d", -i);
    gdk_draw_string(Mixpixmap, fnt, gc, 2, MINY+i+4, str);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, MINXI, MINY+i, Grwidth-MINXI, 1);
  }
  gdk_draw_rectangle(Mixpixmap, gc, TRUE, MINXI, MINY+128, Grwidth-MINXI, 1);
*/
  x=0;

  // Highlight
  gdk_gc_set_foreground(gc, &Hilight);
  if (InputChannel>=0)
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*InputChannel, XCELLTOT*(OutputChannel+1), YCELLTOT);
//    gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*InputChannel, Mixwidth, YCELLTOT);
  if (OutputChannel>=0)
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XCELLTOT*(OutputChannel+1), YCELLTOT*InputChannel, XCELLTOT, Mixheight);
//    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XCELLTOT*(OutputChannel+1), 0, XCELLTOT, Mixheight);

  // Draw the grid
  gdk_gc_set_font(gc, fnt);
  gdk_gc_set_foreground(gc, &Grid);
  for (i=0; i<nIn; i++) {
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, 0, YCELLTOT*(i+1)-1, Mixwidth, 1);
    if (i<fdIn)
      sprintf(str, "A%d", i);
    else
      sprintf(str, "D%d", i-fdIn);
    gdk_draw_string(Mixpixmap, fnt, gc, 1, YCELLTOT*i+(YCELLTOT>>1)+4, str);
  }
  for (o=0; o<nOut; o++) {
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XCELLTOT*(o+1), 0, 1, Mixheight);
    if (o<fdOut)
      sprintf(str, "A%d", o);
    else
      sprintf(str, "D%d", o-fdOut);
    gdk_draw_string(Mixpixmap, fnt, gc, XCELLTOT*(o+1)+2, YCELLTOT*nIn+(YCELLTOT>>1)+4, str);
  }
  gdk_draw_string(Mixpixmap, fnt, gc, 1, YCELLTOT*nIn+8, "In");
  gdk_draw_string(Mixpixmap, fnt, gc, XCELLTOT-18, YCELLTOT*(nIn+1)-2, "Out");


  // Draw input levels and peaks
  for (i=0; i<inchannels; i++) {
    dB=InLevel[i]>>8;
    db=dB>>2;
    gdk_gc_set_foreground(gc, &Bars);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
    if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
      gdk_gc_set_foreground(gc, &Bars1);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db-1, BARWIDTH, 1);
    }
  }
  gdk_gc_set_foreground(gc, &Peak);
  for (i=0; i<inchannels; i++) {
    db=InPeak[i]>>10;
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER, YCELLTOT*i+YCELLBORDER-db, BARWIDTH, 1);
  }

  // Draw output levels and peaks
  for (o=0; o<outchannels; o++) {
    dB=OutLevel[o]>>8;
    db=dB>>2;
    gdk_gc_set_foreground(gc, &Bars);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*nIn+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
    if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
      gdk_gc_set_foreground(gc, &Bars1);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*nIn+YCELLBORDER-db-1, BARWIDTH, 1);
    }
    db=OutPeak[o]>>10;
    gdk_gc_set_foreground(gc, &Peak);
    gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*nIn+YCELLBORDER-db, BARWIDTH, 1);
  }

  for (o=0; o<outchannels; o++) {
    for (i=0; i<inchannels; i++) {
      dB=(Mixer[o][i].val+InLevel[i])>>8;
      db=dB>>2;
      if (db<-YCELLDIM)
        db=-YCELLDIM;
      gdk_gc_set_foreground(gc, &Bars);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db, BARWIDTH, YCELLDIM+db);
      if (dB>-(YCELLDIM<<2))
        if (Bars1.pixel=(((dB&3)*0x40)<<8)) {
          gdk_gc_set_foreground(gc, &Bars1);
          gdk_draw_rectangle(Mixpixmap, gc, TRUE, XMETER+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db-1, BARWIDTH, 1);
        }
      gdk_gc_set_foreground(gc, &Level);
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER, 1, YCELLDIM);
      db=Mixer[o][i].val>>10;
      gdk_draw_rectangle(Mixpixmap, gc, TRUE, XVOLUME-2+XCELLTOT*(o+1), YCELLTOT*i+YCELLBORDER-db, 5, 1);
    }
  }

  gtk_widget_draw(Mixdarea, &update_rect);
  return(TRUE);
}



// Draw the VU-meter
gint UpdateVUmeters(gpointer data) {
  GdkRectangle update_rect;
  int InLevel[ECHO_MAXAUDIOINPUTS];
  int InPeak[ECHO_MAXAUDIOINPUTS];
  int OutLevel[ECHO_MAXAUDIOOUTPUTS];
  int OutPeak[ECHO_MAXAUDIOOUTPUTS];
  static int InClip[ECHO_MAXAUDIOINPUTS];
  static int OutClip[ECHO_MAXAUDIOOUTPUTS];
  int i, x, dB;
  char str[16];
  GdkColor Selected={0xC86060, 0, 0, 0};
  GdkColor Grid={0x9694C4, 0, 0, 0};
  GdkColor Grid2={0x646383, 0, 0, 0};
  GdkColor dBValues={0x00B000, 0, 0, 0};
  GdkColor AnBars={0x00E0B8, 0, 0, 0};
  GdkColor DiBars={0x98E000, 0, 0, 0};
  GdkColor ClipPeak={0, 0, 0, 0};
  GdkColor Peak={0x00FF00, 0, 0, 0};

#define MINXI 30
#define MINXO 132
#define MINY 20

  if (!VUpixmap)
    return(TRUE);

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = VUwidth;
  update_rect.height = VUheight;
  GetVUmeters(InLevel, InPeak, OutLevel, OutPeak);

  if (!gc) {
    gc=gdk_gc_new(gtk_widget_get_parent_window(VUdarea));
    for (i=0; i<nIn; i++)
      InClip[i]=0;
    for (i=0; i<nOut; i++)
      OutClip[i]=0;
  }

  // Clear the image
  gdk_draw_rectangle(VUpixmap, VUdarea->style->black_gc, TRUE, 0, 0, VUwidth, VUheight);

  // Draw the dB scale and the grid
  gdk_gc_set_font(gc, fnt);
  gdk_gc_set_foreground(gc, &Peak);
  gdk_draw_string(VUpixmap, fnt, gc, 2, MINY-12+4, "  dB");
  for (i=0; i<=120; i+=12) {
    sprintf(str, "%4d", -i);
    gdk_gc_set_foreground(gc, &dBValues);
    gdk_draw_string(VUpixmap, fnt, gc, 2, MINY+i+4, str);
    gdk_gc_set_foreground(gc, &Grid);
    gdk_draw_rectangle(VUpixmap, gc, TRUE, MINXI, MINY+i, VUwidth-MINXI, 1);
  }
  gdk_gc_set_foreground(gc, &Grid2);
  gdk_draw_rectangle(VUpixmap, gc, TRUE, MINXI, MINY+128, VUwidth-MINXI, 1);

  x=MINXI+BARSEP;

  for (i=0; i<nIn; i++) {
    if (i<fdIn)
      gdk_gc_set_foreground(gc, &AnBars);
    else
      gdk_gc_set_foreground(gc, &DiBars);
    dB=InLevel[i]>>8;
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 129+MINY-(MINY-dB));

    dB=InPeak[i]>>8;
    if (dB==0)
      InClip[i]=64;
    if (InClip[i]) {
      InClip[i]--;
      ClipPeak.pixel=(InClip[i]<<18)+((255-(InClip[i]*3))<<8);
      gdk_gc_set_foreground(gc, &ClipPeak);
    } else {
      gdk_gc_set_foreground(gc, &Peak);
    }
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 1);
    if (InputChannel==i) {
      gdk_gc_set_foreground(gc, &Selected);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x+1, MINY+128+3, BARWIDTH-2, 1);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY+128+4, BARWIDTH, 1);
    }
    x+=BARWIDTH+BARSEP;
  }

  x+=BARWIDTH+BARSEP;
  for (i=0; i<nOut; i++) {
    if (i<fdOut)
      gdk_gc_set_foreground(gc, &AnBars);
    else
      gdk_gc_set_foreground(gc, &DiBars);
    dB=OutLevel[i]>>8;
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 129+MINY-(MINY-dB));

    dB=OutPeak[i]>>8;
    if (dB==0)
      OutClip[i]=64;
    if (OutClip[i]) {
      OutClip[i]--;
      ClipPeak.pixel=(OutClip[i]<<18)+((255-(OutClip[i]*3))<<8);
      gdk_gc_set_foreground(gc, &ClipPeak);
    } else {
      gdk_gc_set_foreground(gc, &Peak);
    }
    gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY-dB, BARWIDTH, 1);
    if (OutputChannel==i) {
      gdk_gc_set_foreground(gc, &Selected);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x+1, MINY+128+3, BARWIDTH-2, 1);
      gdk_draw_rectangle(VUpixmap, gc, TRUE, x, MINY+128+4, BARWIDTH, 1);
    }
    x+=BARWIDTH+BARSEP;
  }
  gtk_widget_draw(VUdarea, &update_rect);

  return(TRUE);
}



///////////////////// GUI events


// Create a new backing pixmap of the appropriate size
static gint VU_configure_event(GtkWidget *widget, GdkEventConfigure *event) {
  if (VUpixmap)
    gdk_pixmap_unref(VUpixmap);
  VUpixmap=gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle(VUpixmap, widget->style->black_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  return(TRUE);
}



// Redraw the screen from the backing pixmap
static gint VU_expose(GtkWidget *widget, GdkEventExpose *event) {
  if (VUpixmap)
    gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], VUpixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);
  return(FALSE);
}



// Create a new backing pixmap of the appropriate size
static gint Mixer_configure_event(GtkWidget *widget, GdkEventConfigure *event) {
  if (Mixpixmap)
    gdk_pixmap_unref(Mixpixmap);
  Mixpixmap=gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
  gdk_draw_rectangle(Mixpixmap, widget->style->black_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
  return(TRUE);
}



// Redraw the screen from the backing pixmap
static gint Mixer_expose(GtkWidget *widget, GdkEventExpose *event) {
  if (Mixpixmap)
    gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)], Mixpixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);
  return(FALSE);
}



void Input_selector_clicked(GtkWidget *widget, gpointer ich) {
  int err, och, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (InputChannel==(int)ich)
    return;

  InputChannel=(int)ich;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  for (och=0; och<nOut; och++) {
    snd_ctl_elem_id_set_numid(id, Mixer[och][InputChannel].id);
    snd_ctl_elem_value_set_id(control, id);

    if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
      printf("Control %s element read error: %s\n", card, snd_strerror(err));
    }
    Mixer[och][InputChannel].val=val=snd_ctl_elem_value_get_integer(control, 0);
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-val;    // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[och]), (gfloat)val);
  }
}



void Output_selector_clicked(GtkWidget *widget, gpointer och) {
  int err, ich, val;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;

  if (OutputChannel==(int)och)
    return;

  OutputChannel=(int)och;
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  for (ich=0; ich<nIn; ich++) {
    snd_ctl_elem_id_set_numid(id, Mixer[OutputChannel][ich].id);
    snd_ctl_elem_value_set_id(control, id);

    if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
      printf("Control %s element read error: %s\n", card, snd_strerror(err));
    }
    Mixer[OutputChannel][ich].val=val=snd_ctl_elem_value_get_integer(control, 0);
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-val;    // Inverts the value
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[ich]), (gfloat)val);
  }
}




#ifdef REVERSE

static gint Mixer_button_press(GtkWidget *widget, GdkEventButton *event) {
  int ch;

  ch=event->x/XCELLTOT-1;
  if (ch!=OutputChannel) {
    if (ch<nOut && (ADATmode() || ch<fdOut+2)) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixout[ch]), TRUE);
      //OutputChannel=ch;
    }
  }
  ch=event->y/YCELLTOT;
  if (ch!=InputChannel) {
    if (ch<nIn && (ADATmode() || ch<fdIn+2)) {
      gtk_widget_grab_focus(GTK_WIDGET(mixvol[ch]));
      //InputChannel=ch;
    }
  }

  if (event->button) {
    mouseY=event->y;
  }
  return TRUE;
}

#else //REVERSE

static gint Mixer_button_press(GtkWidget *widget, GdkEventButton *event) {
  int ch;

  ch=event->y/YCELLTOT;
  if (ch!=InputChannel) {
    if (ch<nIn && (ADATmode() || ch<fdIn+2)) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mixin[ch]), TRUE);
      //InputChannel=ch;
    }
  }
  ch=event->x/XCELLTOT-1;
  if (ch!=OutputChannel) {
    if (ch<nOut && (ADATmode() || ch<fdOut+2)) {
      gtk_widget_grab_focus(GTK_WIDGET(mixvol[ch]));
      //OutputChannel=ch;
    }
  }

  if (event->button) {
    mouseY=event->y;
  }
  return(TRUE);
}

#endif //REVERSE



static gint Mixer_motion_notify(GtkWidget *widget, GdkEventMotion *event) {
  int x, y;
  GdkModifierType state;
  float val;

  if (event->is_hint)
    gdk_window_get_pointer(event->window, &x, &y, &state);
  else {
    x=event->x;
    y=event->y;
    state=event->state;
  }

  if (state & GDK_BUTTON1_MASK && Mixpixmap != NULL) {
    if (!mouseButton)
      mouseButton=1;
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-Mixer[OutputChannel][InputChannel].val;          // Invert
    val+=256*(y-mouseY);
    mouseY=y;
    // Gtk limits the range of "val"
#ifdef REVERSE
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[InputChannel]), (gfloat)val);
#else
    gtk_adjustment_set_value(GTK_ADJUSTMENT(mixeradj[OutputChannel]), (gfloat)val);
#endif
  } else
    mouseButton=0;

  return(TRUE);
}




gint Mainwindow_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {
  if (VUwindow)
    gtk_widget_destroy(VUwindow);
  gtk_main_quit();
  return(FALSE);
}



gint VUwindow_destroy(GtkWidget *widget, gpointer data) {
  SetVUmeters(0);
  gtk_timeout_remove(VUtimer);
  //@@del gc and fnt
  VUwindow=0;
  return(TRUE);
}



gint Mixwindow_destroy(GtkWidget *widget, gpointer data) {
  SetVUmeters(0);
  gtk_timeout_remove(Mixtimer);
  //@@del gc and fnt
  Mixwindow=0;
  return(TRUE);
}



void Monitor_volume_changed(GtkWidget *widget, gpointer cnl) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  //unsigned int count;
  int err, val, ch;
  char str[16];
  struct mixel *mxl;
  //snd_ctl_elem_type_t type;
  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value

  ch=(int)cnl;
  if (ch<ECHO_MAXAUDIOINPUTS)
    mxl=&Mixer[OutputChannel][ch];
  else {
    ch-=ECHO_MAXAUDIOINPUTS;
    mxl=&Mixer[ch][InputChannel];
  }

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_id_set_numid(id, mxl->id);
/*
  snd_ctl_elem_info_set_id(info, id);
  if ((err = snd_ctl_elem_info(handle, info)) < 0) {
    printf("Control %s cinfo error: %s\n", card, snd_strerror(err));
  }
  snd_ctl_elem_info_get_id(info, id);
  type = snd_ctl_elem_info_get_type(info);
  count = snd_ctl_elem_info_get_count(info);
//printf("%s\n", snd_ctl_elem_info_get_name(info));
*/
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, val);
  if ((err = snd_ctl_elem_write(ctlhandle, control)) < 0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  } else {
    mxl->val=val;
    if (val==ECHOGAIN_MINOUT)
      strcpy(str, "mute");
    else
      sprintf(str, "%+3.0f", (float)val/256.0);
    gtk_label_set_text(GTK_LABEL(mixvol_lbl[ch]), str);
  }
}



void Monitor_outvolume_clicked(GtkWidget *widget, gpointer ch) {

  OutputChannel=(int)ch;
}



void Monitor_involume_clicked(GtkWidget *widget, gpointer ch) {

  InputChannel=(int)ch;
}



void Analog_volume_changed(GtkWidget *widget, gpointer ch) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  char str[16], *fmt;
  int err, channel, val;
  GtkWidget **label;
  GtkObject **adj;

  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  if ((int)ch<ECHO_MAXAUDIOINPUTS) {
    channel=(int)ch;
    label=invol_lbl;
    fmt="%+4.1f";
    adj=involadj;
    val=ECHOGAIN_MININP+ECHOGAIN_MAXINP-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value
    snd_ctl_elem_id_set_numid(id, involId);
  } else {
    channel=(int)ch-ECHO_MAXAUDIOINPUTS;
    label=outvol_lbl;
    fmt="%+3.0f";
    adj=outvoladj;
    val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value
    snd_ctl_elem_id_set_numid(id, outvolId);
  }

  if (val==ECHOGAIN_MINOUT)
    strcpy(str, "mute");
  else
    sprintf(str, fmt, (float)val/256.0);
  gtk_label_set_text(GTK_LABEL(label[channel]), str);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_value_set_id(control, id);

  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  snd_ctl_elem_value_set_integer(control, channel, val);
  if (Gang)
    snd_ctl_elem_value_set_integer(control, channel^0x01, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
  if (Gang)
    gtk_adjustment_set_value(GTK_ADJUSTMENT(adj[channel^0x01]), (gfloat)GTK_ADJUSTMENT(widget)->value);
}



void Vmixer_volume_changed(GtkWidget *widget, gpointer idx) {
  snd_ctl_elem_info_t *info;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  char str[16];
  int err, val, pipe, output, index;

  snd_ctl_elem_info_alloca(&info);
  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  index=(int)idx;
  pipe=index>>1;
  output=index&1;

  val=ECHOGAIN_MINOUT+ECHOGAIN_MAXOUT-(int)GTK_ADJUSTMENT(widget)->value;    // Inverts the value

  if (val==ECHOGAIN_MINOUT)
    strcpy(str, "mute");
  else
    sprintf(str, "%+.0f", (float)val/256.0);
  gtk_label_set_text(GTK_LABEL(vmixer_lbl[index]), str);

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  snd_ctl_elem_value_set_id(control, id);

  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  snd_ctl_elem_value_set_integer(control, pipe*nVOut+output, val);
  if (Gang)
    snd_ctl_elem_value_set_integer(control, (pipe*nVOut+output)^0x01, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
  if (Gang)
    gtk_adjustment_set_value(GTK_ADJUSTMENT(vmixeradj[index^0x01]), (gfloat)GTK_ADJUSTMENT(widget)->value);
}



void Nominal_level_toggled(GtkWidget *widget, gpointer ch) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val, item;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  val=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_MIXER);
  if ((int)ch<ECHO_MAXAUDIOINPUTS) {
    item=(int)ch;
    snd_ctl_elem_id_set_numid(id, p4InId);
  } else {
    item=(int)ch-ECHO_MAXAUDIOINPUTS;
    snd_ctl_elem_id_set_numid(id, p4OutId);
  }
  snd_ctl_elem_value_set_id(control, id);
  if ((err=snd_ctl_elem_read(ctlhandle, control))<0) {
    printf("Control %s element read error: %s\n", card, snd_strerror(err));
  }
  snd_ctl_elem_value_set_integer(control, item, !val);  // FALSE is +4
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}


void Dithering_toggled(GtkWidget *widget, gpointer g) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err, val;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);

  val=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_CARD);
  snd_ctl_elem_id_set_numid(id, ditherId);
  snd_ctl_elem_value_set_id(control, id);
  snd_ctl_elem_value_set_integer(control, 0, val);
  if ((err=snd_ctl_elem_write(ctlhandle, control))<0) {
    printf("Control %s element write error: %s\n", card, snd_strerror(err));
  }
}



void VUmeters_button_click(GtkWidget *widget, gpointer n) {

  if (!VUwindow) {
    // Create VU-meter window
    VUwidth=MINXI+(BARWIDTH+BARSEP)*(nIn+nOut+1)+BARSEP;
    VUheight=160;
    SetVUmeters(1);
    VUwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (VUwindow), "VU-meters");
    gtk_signal_connect(GTK_OBJECT(VUwindow), "destroy", GTK_SIGNAL_FUNC(VUwindow_destroy), NULL);
    gtk_window_set_policy(GTK_WINDOW(VUwindow), FALSE, FALSE, TRUE);
    gtk_widget_show(VUwindow);

    VUdarea=gtk_drawing_area_new();
    gtk_widget_set_events(VUdarea, GDK_EXPOSURE_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
    gtk_drawing_area_size(GTK_DRAWING_AREA(VUdarea), VUwidth, VUheight);
    gtk_container_add(GTK_CONTAINER(VUwindow), VUdarea);

    gtk_widget_show(VUdarea);
    gtk_signal_connect(GTK_OBJECT(VUdarea), "expose_event", (GtkSignalFunc)VU_expose, NULL);
    gtk_signal_connect(GTK_OBJECT(VUdarea), "configure_event", (GtkSignalFunc)VU_configure_event, NULL);
    VUtimer=gtk_timeout_add(30, UpdateVUmeters, 0);  // The hw updates the meters about 30 times/s
    gdk_window_clear_area(VUdarea->window, 0, 0, VUwidth, VUheight);
  }
}



void Mixer_button_click(GtkWidget *widget, gpointer n) {

  if (!Mixwindow) {
    // Create VU-meter window
    Mixwidth=XCELLTOT*(nOut+1);
    Mixheight=YCELLTOT*(nIn+1);
    SetVUmeters(1);
    Mixwindow=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (Mixwindow), "Mixer");
    gtk_signal_connect(GTK_OBJECT(Mixwindow), "destroy", GTK_SIGNAL_FUNC(Mixwindow_destroy), NULL);
    gtk_window_set_policy(GTK_WINDOW(Mixwindow), FALSE, FALSE, TRUE);
    gtk_widget_show(Mixwindow);

    Mixdarea=gtk_drawing_area_new();
    gtk_widget_set_events(Mixdarea, GDK_EXPOSURE_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
    gtk_drawing_area_size(GTK_DRAWING_AREA(Mixdarea), Mixwidth, Mixheight);
    gtk_container_add(GTK_CONTAINER(Mixwindow), Mixdarea);

    gtk_widget_show(Mixdarea);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "expose_event", (GtkSignalFunc)Mixer_expose, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "configure_event", (GtkSignalFunc)Mixer_configure_event, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "motion_notify_event", (GtkSignalFunc)Mixer_motion_notify, NULL);
    gtk_signal_connect(GTK_OBJECT(Mixdarea), "button_press_event", (GtkSignalFunc)Mixer_button_press, NULL);
    Mixtimer=gtk_timeout_add(30, UpdateMixer, 0);  // The hw updates the meters about 30 times/s
    gdk_window_clear_area(Mixdarea->window, 0, 0, Mixwidth, Mixheight);
  }
}



void Gang_button_toggled(GtkWidget *widget, gpointer x) {

  Gang=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}



void Digital_mode_activate(GtkWidget *widget, gpointer mode) {

  if (SetEnum(dmodeId, (int)mode)<0)     // When I change the digital mode, the clock source can change too
    gtk_option_menu_set_history(GTK_OPTION_MENU(dmodeOpt), dmodeVal);
  else {
    dmodeVal=(int)mode;
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal=GetEnum(clocksrcId));
  }
  SetSensitivity(ADATmode());
}



void Clock_source_activate(GtkWidget *widget, gpointer clk) {

  if (SetEnum(clocksrcId, (int)clk)<0)
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal);
  else
    clocksrcVal=(int)clk;
}



void SPDIF_mode_activate(GtkWidget *widget, gpointer mode) {

  SetEnum(spdifmodeId, (int)mode);         // This one should never fail
  spdifmodeVal=(int)mode;
}



/*
int GetMaxChannels(int pcmnum, int inout) {
snd_pcm_t *pcm_handle;         
snd_pcm_stream_t stream;
snd_pcm_hw_params_t *hwparams;
char pcm_name[64];
int err, n;

  snd_pcm_hw_params_alloca(&hwparams);
  stream=inout;
  sprintf(pcm_name, "%s,%d", card, pcmnum);
  if ((err=snd_pcm_open(&pcm_handle, pcm_name, stream, SND_PCM_NONBLOCK))<0) {
    if (err==-ENOENT)
      return(0);     // Ok, this device does not exists on this card.
    fprintf(stderr, "Error opening PCM device %s (%s)\n", pcm_name, snd_strerror(err));
    exit(1);
  }
  // Init hwparams with full configuration space
  if ((err=snd_pcm_hw_params_any(pcm_handle, hwparams))<0) {
    fprintf(stderr, "Can not configure this PCM device (%s)\n", snd_strerror(err));
    exit(1);
  }
  n=snd_pcm_hw_params_get_channels_max(hwparams);
  snd_pcm_close(pcm_handle);
  return(n);
}
*/



void GetChannels(void) {
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_value_t *control;
  int err;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_value_alloca(&control);
  snd_ctl_elem_id_set_interface(id, SND_CTL_ELEM_IFACE_HWDEP);

  snd_ctl_elem_id_set_numid(id, channelsId);
  snd_ctl_elem_value_set_id(control, id);
  if ((err = snd_ctl_elem_read(ctlhandle, control)) < 0) {
    printf("GetChannels() read error: %s\n", snd_strerror(err));
    exit(1);
  }
  nIn=snd_ctl_elem_value_get_integer(control, 0);      // Number of input channels
  fdIn=snd_ctl_elem_value_get_integer(control, 1);     // First digital in (= number of analog input channels)
  nOut=snd_ctl_elem_value_get_integer(control, 2);     // Number of output channels
  fdOut=snd_ctl_elem_value_get_integer(control, 3);    // First digital out
  nVOut=snd_ctl_elem_value_get_integer(control, 4);    // Number of virtual output channels (==nOut on non-vmixer cards)
}



// Scan all controls and store their ID's for later use.
int OpenControls(const char *card, const char *cardname) {
  int err, i, o;
  int FirstMixerId, FirstVmixerId, numid, count, items, item;
  snd_hctl_t *handle;
  snd_hctl_elem_t *elem;
  snd_ctl_elem_id_t *id;
  snd_ctl_elem_info_t *info;

  outvolId=masterId=vmixerId=p4InId=p4OutId=dmodeId=clocksrcId=spdifmodeId=vuswitchId=vumetersId=0;
  outvolCount=0;

  snd_ctl_elem_id_alloca(&id);
  snd_ctl_elem_info_alloca(&info);

  ndmodes=nclocksrc=nspdifmodes=0;

  FirstMixerId=FirstVmixerId=-1;
  if ((err=snd_hctl_open(&handle, card, 0))<0) {
    printf("Control %s open error: %s", card, snd_strerror(err));
    return err;
  }
  if ((err=snd_hctl_load(handle))<0) {
    printf("Control %s local error: %s\n", card, snd_strerror(err));
    return err;
  }
  for (elem=snd_hctl_first_elem(handle); elem; elem=snd_hctl_elem_next(elem)) {
    if ((err=snd_hctl_elem_info(elem, info))<0) {
      printf("Control %s snd_hctl_elem_info error: %s\n", card, snd_strerror(err));
      return err;
    }
    if (snd_ctl_elem_info_is_inactive(info))
      continue;
    snd_hctl_elem_get_id(elem, id);
    numid=snd_ctl_elem_id_get_numid(id);
    count=snd_ctl_elem_info_get_count(info);
    if (!strcmp("Monitor Mixer Volume", snd_ctl_elem_id_get_name(id))) {
      if (FirstMixerId<0) {
        FirstMixerId=numid;
printf("First Mixer id=%d\n", FirstMixerId);
        mixerOutputs=snd_ctl_elem_info_get_dimension(info, 0);
        mixerInputs=snd_ctl_elem_info_get_dimension(info, 1);
      }
    } else if (!strcmp("VMixer Volume", snd_ctl_elem_id_get_name(id))) {
      if (FirstVmixerId<0) {
        FirstVmixerId=numid;
printf("First Vmixer id=%d\n", FirstVmixerId);
        vmixerPipes=snd_ctl_elem_info_get_dimension(info, 0);
        vmixerOutputs=snd_ctl_elem_info_get_dimension(info, 1);
      }
    } else if (!strcmp("PCM Playback Volume", snd_ctl_elem_id_get_name(id))) {
      outvolId=snd_ctl_elem_id_get_numid(id);
      outvolCount=count;
printf("Playback Volume id=%d [%d]\n", outvolId, count);
    } else if (!strcmp("Master Playback Volume", snd_ctl_elem_id_get_name(id))) {
      masterId=snd_ctl_elem_id_get_numid(id);
printf("Master Volume id=%d\n", masterId);
    } else if (!strcmp("PCM Capture Volume", snd_ctl_elem_id_get_name(id))) {
      involId=snd_ctl_elem_id_get_numid(id);
printf("Capture Volume id=%d [%d]\n", involId, count);
    } else if (!strcmp("Line Playback Switch (-10dBV)", snd_ctl_elem_id_get_name(id))) {
      p4OutId=numid;
printf("Playback nominal id=%d [%d]\n", p4OutId, count);
    } else if (!strcmp("Line Capture Switch (-10dBV)", snd_ctl_elem_id_get_name(id))) {
      p4InId=numid;
printf("Capture nominal id=%d [%d]\n", p4InId, count);
    } else if (!strcmp("Digital Mode Switch", snd_ctl_elem_id_get_name(id))) {
      dmodeId=numid;
      items=snd_ctl_elem_info_get_items(info);
      ndmodes=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info)) < 0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(dmodeName[item], snd_ctl_elem_info_get_item_name(info), 63);
        dmodeName[item][63]=0;
printf("Digital Mode id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("Clock source Switch", snd_ctl_elem_id_get_name(id))) {
      clocksrcId=numid;
      items=snd_ctl_elem_info_get_items(info);
      nclocksrc=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info))<0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(clocksrcName[item], snd_ctl_elem_info_get_item_name(info), 63);
        clocksrcName[item][63]=0;
printf("Clock source id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("S/PDIF Mode Switch", snd_ctl_elem_id_get_name(id))) {
      spdifmodeId=numid;
      items=snd_ctl_elem_info_get_items(info);
      nspdifmodes=items;
      for (item=0; item<items; item++) {
        snd_ctl_elem_info_set_item(info, item);
        if ((err=snd_hctl_elem_info(elem, info)) < 0) {
          printf("Control %s element info error: %s\n", card, snd_strerror(err));
        }
        strncpy(spdifmodeName[item], snd_ctl_elem_info_get_item_name(info), 63);
        spdifmodeName[item][63]=0;
printf("S/PDIF Mode id=%d item #%u '%s'\n", numid, item, snd_ctl_elem_info_get_item_name(info));
      }
    } else if (!strcmp("Dithering Switch", snd_ctl_elem_id_get_name(id))) {
      ditherId=numid;
printf("Dithering switch id=%d\n", numid);
    } else if (!strcmp("VU-meters Switch", snd_ctl_elem_id_get_name(id))) {
      vuswitchId=numid;
printf("VU-meter switch id=%d\n", numid);
    } else if (!strcmp("VU-meters", snd_ctl_elem_id_get_name(id))) {
      vumetersId=numid;
printf("VU-meters id=%d\n", numid);
    } else if (!strcmp("Channels info", snd_ctl_elem_id_get_name(id))) {
      channelsId=numid;
printf("Channels info id=%d\n", numid);
    }
  }

  GetChannels();
printf("nIn=%d fdIn=%d nOut=%d nVOut=%d fdOut=%d\n", nIn,fdIn,nOut,nVOut, fdOut);

  if (FirstMixerId>=0 && (mixerInputs!=nIn || mixerOutputs!=nOut)) {
    printf("** Warning ** Mixer/channels mismatch !!  nIn=%d mIn=%d    nOut=%d mOut=%d\n", nIn, mixerInputs, nOut, mixerOutputs);
    return(1);
  }
  if (FirstVmixerId>=0) {
    if (vmixerPipes!=nVOut || vmixerOutputs!=fdOut) {
      printf("error: vmixer/channels mismatch  vmp=%d nvo=%d    vmo=%d fdo=%d !!\n", vmixerPipes, nVOut, vmixerOutputs, fdOut);
      return(1);
    }
    if (vmixerOutputs>2) {
      printf("error: more than 2 vmixer outputs are not supported.\n");
      return(1);
    }
  }
  
  //@ Assumes all mixer and vmixer controls are contiguous
  for (o=0; o<nOut; o++) {
    for (i=0; i<nIn; i++) {
      Mixer[o][i].id=FirstMixerId++;
    }
  }

  for (i=0; i<nOut; i++) {
    for (o=0; o<nVOut; o++) {
      Vmixer[i][o].id=FirstVmixerId++;
    }
  }

  // For compatibility with the Mia.
  if (masterId)
    outvolId=masterId;

  snd_hctl_close(handle);
  return(0);
}



int main(int argc, char *argv[]) {
  gchar str[256];
  GtkWidget *hbox, *vbox;
  GtkWidget *vbmain, *hbmain1, *hbmain2;
  GtkWidget *vbsel, *frame, *button;
  GtkWidget *label, *menu, *menuitem;
  GSList *in_wgroup, *out_wgroup;
  int err, i, n, cardnum;
  char name[8], cardname[32];
  snd_ctl_card_info_t *hw_info;


  // Scans all installed cards
  snd_ctl_card_info_alloca(&hw_info);
  cardnum=-1;
  ctlhandle=0;

  while (snd_card_next(&cardnum)>=0 && cardnum>=0) {
    sprintf(name, "hw:%d", cardnum);
    if ((err=snd_ctl_open(&ctlhandle, name, 0))<0) {
      printf("snd_ctl_open(%s) Error: %s\n", name, snd_strerror(err));
      continue;
    }
    if ((err=snd_ctl_card_info(ctlhandle, hw_info))>=0) {
      if (!strcmp(snd_ctl_card_info_get_driver(hw_info), "Echoaudio")) {
        strcpy(card, name);
        strcpy(cardname, snd_ctl_card_info_get_name(hw_info));
        printf("Card found: %s  (%s)\n", snd_ctl_card_info_get_longname(hw_info), name);
        break;
      }
    } else {
      printf("snd_ctl_card_info(%s) Error: %s\n", name, snd_strerror(err));
    }
    snd_ctl_close(ctlhandle);
    ctlhandle=0;
  }

  if (!ctlhandle) {
    printf("No Echoaudio cards found, sorry.\n");
    return(0);
  }

  // Reads available controls
  if (OpenControls(card, cardname))
    exit(1);

  mouseButton=0;
  Gang=1;

  gtk_init(&argc, &argv);
  fnt=gdk_font_load("-misc-fixed-medium-r-*-*-10-*-*-*-*-*-*-*");
  if (!fnt) {
    printf("Cannot find the font\n");
    exit(1);
  }

/*
 -window-------------------------------------------------------------------
|                                                                          |
|  -vbox vbmain----------------------------------------------------------  |
| |                                                                      | |
| |  -hbox hbmain1-----------------------------------------------------  | |
| | |                                                                  | | |
| | |  -frame---------                                                 | | |
| | | |               | ....                                           | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | | |               |                                                | | |
| | |  ---------------                                                 | | |
| | |                                                                  | | |
| |  ------------------------------------------------------------------  | |
| |  -hbox hbmain2-----------------------------------------------------  | |
| | |                                                                  | | |
| | |                                                                  | | |
| | |                                                                  | | |
| | |                                                                  | | |
| |  ------------------------------------------------------------------  | |
| |                                                                      | |
|  ----------------------------------------------------------------------  |
|                                                                          |
 --------------------------------------------------------------------------
*/

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), VERSION);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(Mainwindow_delete), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(window), BORDER);

  vbmain=gtk_vbox_new(FALSE, SPACING);
  gtk_widget_show(vbmain);
  gtk_container_add(GTK_CONTAINER(window), vbmain);

///////////////////////////// Upper row

  hbmain1=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain1);
  gtk_box_pack_start(GTK_BOX(vbmain), hbmain1, TRUE, TRUE, 0);


  // Input volume widgets
  if (involId) {
    frame=gtk_frame_new("Analog input volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdIn; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      sprintf(str, "%d", i);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      involadj[i]=gtk_adjustment_new(0, ECHOGAIN_MININP, ECHOGAIN_MAXINP, LEVELSTEP/2, LEVELSTEP*3, 0);
      invol[i]=gtk_vscale_new(GTK_ADJUSTMENT(involadj[i]));
      gtk_widget_show(invol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), invol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(invol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(involadj[i]), "value_changed", GTK_SIGNAL_FUNC(Analog_volume_changed), (gpointer)i);
      // Value label
      invol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(invol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), invol_lbl[i], FALSE, FALSE, 0);
    }
    InitVolumeGUI(INPUT);
  }


  // Output volume widgets
  if (outvolId) {
    frame=gtk_frame_new("Output volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<outvolCount; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdOut)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdOut);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      outvoladj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      outvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(outvoladj[i]));
      gtk_widget_show(outvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), outvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(outvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(outvoladj[i]), "value_changed", GTK_SIGNAL_FUNC(Analog_volume_changed), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
      // Value label
      outvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(outvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), outvol_lbl[i], FALSE, FALSE, 0);
    }
    InitVolumeGUI(OUTPUT);
  }

/*
  // Vmixer volume widgets
  if (Vmixer[0][0].id>=0) {
    frame=gtk_frame_new("Vmixer volume");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nVOut*fdOut; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i&1)
        sprintf(str, "%dR", i>>1);
      else
        sprintf(str, "%dL", i>>1);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      vmixeradj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      vmixervol[i]=gtk_vscale_new(GTK_ADJUSTMENT(vmixeradj[i]));
      gtk_widget_show(vmixervol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), vmixervol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(vmixervol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(vmixeradj[i]), "value_changed", GTK_SIGNAL_FUNC(Vmixer_volume_changed), (gpointer)i);
      // Value label
      vmixer_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(vmixer_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), vmixer_lbl[i], FALSE, FALSE, 0);
    }
//    InitVolumeGUI(OUTPUT);
  }
*/
#warning Vmixer_volume_changed

  // Mixer
  if (Mixer[0][0].id>=0) {
#ifdef REVERSE
    // Mixer volume widgets
    frame=gtk_frame_new("Mixer input levels");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nIn; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdIn)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdIn);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      mixeradj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      mixvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(mixeradj[i]));
      gtk_widget_show(mixvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(mixvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(mixvol[i]), "grab_focus", GTK_SIGNAL_FUNC(Monitor_involume_clicked), (gpointer)i);
      gtk_signal_connect(GTK_OBJECT(mixeradj[i]), "value_changed", GTK_SIGNAL_FUNC(Monitor_volume_changed), (gpointer)i);
      // Value label
      mixvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(mixvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol_lbl[i], FALSE, FALSE, 0);
    }
    InputChannel=OutputChannel=-1;

    // Output channel selectors
    frame=gtk_frame_new("Mixer output");
    gtk_widget_show(frame);
    vbsel=gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, FALSE, FALSE, 0);
    gtk_widget_show(vbsel);
    gtk_container_add(GTK_CONTAINER(frame), vbsel);

    out_wgroup=0;
    for (i=n=0; i<nOut; i++) {
      if (i<fdOut)
        sprintf(str, "An-%d", i);
      else
        sprintf(str, "Di-%d", i-fdOut);
      if (i)
        out_wgroup=gtk_radio_button_group(GTK_RADIO_BUTTON(mixout[i-1]));
      mixout[i]=gtk_radio_button_new_with_label(out_wgroup, str);
      gtk_widget_show(mixout[i]);
      gtk_box_pack_start(GTK_BOX(vbsel), mixout[i], FALSE, FALSE, 0);
      gtk_signal_connect(GTK_OBJECT(mixout[i]), "toggled", GTK_SIGNAL_FUNC(Output_selector_clicked), (gpointer)i);
    }
    Output_selector_clicked(0, 0);

#else // REVERSE

    // Input channel selectors
    frame=gtk_frame_new("Mixer input");
    gtk_widget_show(frame);
    vbsel=gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, FALSE, FALSE, 0);
    gtk_widget_show(vbsel);
    gtk_container_add(GTK_CONTAINER(frame), vbsel);

    in_wgroup=0;
    for (i=n=0; i<nIn; i++) {
      if (i<fdIn)
        sprintf(str, "An-%d", i);
      else
        sprintf(str, "Di-%d", i-fdIn);
      if (i)
        in_wgroup=gtk_radio_button_group(GTK_RADIO_BUTTON(mixin[i-1]));
      mixin[i]=gtk_radio_button_new_with_label(in_wgroup, str);
      gtk_widget_show(mixin[i]);
      gtk_box_pack_start(GTK_BOX(vbsel), mixin[i], FALSE, FALSE, 0);
      gtk_signal_connect(GTK_OBJECT(mixin[i]), "toggled", GTK_SIGNAL_FUNC(Input_selector_clicked), (gpointer)i);
    }

    // Mixer volume widgets
    frame=gtk_frame_new("Mixer output levels");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain1), frame, TRUE, TRUE, 0);
    hbox=gtk_hbox_new(TRUE, 1);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<nOut; i++) {
      vbox=gtk_vbox_new(FALSE, 0);
      gtk_widget_show(vbox);
      gtk_container_add(GTK_CONTAINER(hbox), vbox);
      // Channel label
      if (i<fdOut)
        sprintf(str, "A%d", i);
      else
        sprintf(str, "D%d", i-fdOut);
      label=gtk_label_new(str);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
      // Volume
      mixeradj[i]=gtk_adjustment_new(0, ECHOGAIN_MINOUT, ECHOGAIN_MAXOUT, LEVELSTEP, LEVELSTEP*6, 0);
      mixvol[i]=gtk_vscale_new(GTK_ADJUSTMENT(mixeradj[i]));
      gtk_widget_show(mixvol[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol[i], TRUE, TRUE, 0);
      gtk_scale_set_draw_value(GTK_SCALE(mixvol[i]), 0);
      gtk_signal_connect(GTK_OBJECT(mixvol[i]), "grab_focus", GTK_SIGNAL_FUNC(Monitor_outvolume_clicked), (gpointer)i);
      gtk_signal_connect(GTK_OBJECT(mixeradj[i]), "value_changed", GTK_SIGNAL_FUNC(Monitor_volume_changed), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
      // Value label
      mixvol_lbl[i]=gtk_label_new("xxx");
      gtk_widget_show(mixvol_lbl[i]);
      gtk_box_pack_start(GTK_BOX(vbox), mixvol_lbl[i], FALSE, FALSE, 0);
    }
    InputChannel=OutputChannel=-1;
    Input_selector_clicked(0, 0);
#endif
    InitMixer();
  }

///////////////////////////// Bottom row

  hbmain2=gtk_hbox_new(FALSE, SPACING);
  gtk_widget_show(hbmain2);
  gtk_box_pack_start(GTK_BOX(vbmain), hbmain2, FALSE, FALSE, 0);


  if (p4InId) {
    // Consumer/professional analog input switches
    frame=gtk_frame_new("Input +4dBu");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdIn; i++) {
      sprintf(str, "%d", i);
      p4dbuIn[i]=gtk_toggle_button_new_with_label(str);
      gtk_box_pack_start(GTK_BOX(hbox), p4dbuIn[i], TRUE, FALSE, 1);
      gtk_widget_show(p4dbuIn[i]);
      gtk_signal_connect(GTK_OBJECT(p4dbuIn[i]), "toggled", GTK_SIGNAL_FUNC(Nominal_level_toggled), (gpointer)i);
    }
    InitNominalLevelGUI(p4InId);
  }


  if (p4OutId) {
    // Consumer/professional analog output switches
    frame=gtk_frame_new("Output +4dBu");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    for (i=0; i<fdOut; i++) {
      sprintf(str, "%d", i);
      p4dbuOut[i]=gtk_toggle_button_new_with_label(str);
      gtk_box_pack_start(GTK_BOX(hbox), p4dbuOut[i], TRUE, FALSE, 1);
      gtk_widget_show(p4dbuOut[i]);
      gtk_signal_connect(GTK_OBJECT(p4dbuOut[i]), "toggled", GTK_SIGNAL_FUNC(Nominal_level_toggled), (gpointer)(i+ECHO_MAXAUDIOINPUTS));
    }
    InitNominalLevelGUI(p4OutId);
  }


  if (dmodeId && ndmodes>1) {
    // Digital mode switch
    frame=gtk_frame_new("Digital mode");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    dmodeOpt=gtk_option_menu_new();
    gtk_widget_show(dmodeOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<ndmodes; i++) {
      menuitem=gtk_menu_item_new_with_label(dmodeName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", Digital_mode_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(dmodeOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), dmodeOpt, FALSE, FALSE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(dmodeOpt), dmodeVal=GetEnum(dmodeId));
    SetSensitivity(ADATmode());
  }


  if (clocksrcId && nclocksrc>1) {
    // Clock source switch
    frame=gtk_frame_new("Clock source");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    clocksrcOpt=gtk_option_menu_new();
    gtk_widget_show(clocksrcOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<nclocksrc; i++) {
      menuitem=gtk_menu_item_new_with_label(clocksrcName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", Clock_source_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(clocksrcOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), clocksrcOpt, FALSE, FALSE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(clocksrcOpt), clocksrcVal=GetEnum(clocksrcId));
  }


  if (spdifmodeId && nspdifmodes>1) {
    // S/PDIF mode switch
    frame=gtk_frame_new("S/PDIF mode");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    spdifmodeOpt=gtk_option_menu_new();
    gtk_widget_show(spdifmodeOpt);
    menu=gtk_menu_new();
    gtk_widget_show(menu);
    for (i=0; i<nspdifmodes; i++) {
      menuitem=gtk_menu_item_new_with_label(spdifmodeName[i]);
      gtk_widget_show(menuitem);
      gtk_signal_connect(GTK_OBJECT(menuitem), "activate", SPDIF_mode_activate, (gpointer)i);
      gtk_menu_append(GTK_MENU(menu), menuitem);
    }
    gtk_option_menu_set_menu(GTK_OPTION_MENU(spdifmodeOpt), menu);
    gtk_box_pack_start(GTK_BOX(hbox), spdifmodeOpt, TRUE, TRUE, 0);
    gtk_option_menu_set_history(GTK_OPTION_MENU(spdifmodeOpt), spdifmodeVal=GetEnum(spdifmodeId));
  }


  if (ditherId) {
    // 16 bits digital input dithering switch
    frame=gtk_frame_new("Dither");
    gtk_widget_show(frame);
    gtk_box_pack_start(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    ditherToggle=gtk_toggle_button_new_with_label("On");
    gtk_widget_show(ditherToggle);
    gtk_box_pack_start(GTK_BOX(hbox), ditherToggle, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(ditherToggle), "toggled", Dithering_toggled, (gpointer)0);
    InitDitheringGUI(ditherId);
  }


  if (vumetersId && vuswitchId) {
    // VU-meter button
    VUwindow=0;
    frame=gtk_frame_new("VUmeters");
    gtk_widget_show(frame);
    gtk_box_pack_end(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button=gtk_button_new_with_label("Show");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", VUmeters_button_click, 0);
  }


  if (Mixer[0][0].id>=0) {
    // VU-meter button
    VUwindow=0;
    frame=gtk_frame_new("Mixer");
    gtk_widget_show(frame);
    gtk_box_pack_end(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button=gtk_button_new_with_label("Show");
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", Mixer_button_click, 0);
  }


  if (involId || outvolId) {
    // Gang button
    VUwindow=0;
    frame=gtk_frame_new("Gang");
    gtk_widget_show(frame);
    gtk_box_pack_end(GTK_BOX(hbmain2), frame, FALSE, FALSE, 0);
    hbox=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button=gtk_toggle_button_new_with_label("On");
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 1);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "toggled", Gang_button_toggled, 0);
  }


  gtk_widget_show(window);
  gtk_main();

  if (VUwindow) {
    SetVUmeters(0);
    gtk_timeout_remove(VUtimer);
  }
  if (Mixwindow) {
    SetVUmeters(0);
    gtk_timeout_remove(Mixtimer);
  }
  snd_ctl_close(ctlhandle);
  return(0);
}

